"""Routines to check the content of a message as retrieved by
[cattle_grid.testing.features.steps.messaging.receive_message][]"""

from behave import then

from bovine.activitystreams.utils import as_list


@then('The received message contains an URL of mediatype "{media_type}"')  # pyright: ignore[reportCallIssue]
def message_contains_url_of_mediatype(context, media_type):
    """

    ```gherkin
    Then The received message contains an URL of mediatype "text/html"
    ```

    The resolved url is added to `context.media_type_link`.
    """

    message = context.received_object

    urls = as_list(message.get("url", []))

    html_url = next(
        (x for x in urls if isinstance(x, dict) and x.get("mediaType") == media_type),
        None,
    )
    assert html_url

    assert html_url.get("type") == "Link"
    assert html_url.get("href")

    context.media_type_link = html_url.get("href")


@then('The URL resolves to a webpage containing "{text}"')  # pyright: ignore[reportCallIssue]
async def step_impl(context, text):
    """
    ```gherkin
    Then The URL resolves to a webpage
    ```
    """

    async with context.session.get(
        context.media_type_link, headers={"accept": "text/html"}
    ) as response:
        assert response.status == 200, (
            f"Status is {response.status} != 200 for {context.media_type_link}"
        )
        content_type = response.headers["content-type"]
        assert content_type.startswith("text/html"), (
            f"Got response content-type {content_type}"
        )

        content = await response.text()

        assert text in content, f"Content {content} does not contain {text}"



from .globals import global_container, get_engine


async def test_alchemy_database():
    async with global_container.alchemy_database("sqlite+aiosqlite:///:memory:"):
        engine = get_engine()

        assert engine



import pytest
from unittest.mock import AsyncMock

from faststream.rabbit import RabbitBroker, TestRabbitBroker, RabbitQueue

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.dependencies.globals import global_container

from . import create_processing_router


@pytest.fixture
async def mock_subscriber():
    return AsyncMock()


@pytest.fixture
async def broker(mock_subscriber):
    br = RabbitBroker("amqp://guest:guest@localhost:5672/")
    br.include_router(create_processing_router())

    br.subscriber(
        RabbitQueue("test_queue", routing_key="incoming.Activity"),
        exchange=global_container.internal_exchange,
    )(mock_subscriber)

    async with TestRabbitBroker(br, connect_only=False) as tbr:
        yield tbr


async def test_shared_inbox(actor_for_test, broker, mock_subscriber):
    activity = {
        "actor": "http://remote.test/actor",
        "type": "Activity",
        "to": [actor_for_test.actor_id],
    }

    await broker.publish(
        {"data": activity},
        routing_key="shared_inbox",
        exchange=global_container.internal_exchange,
    )

    mock_subscriber.assert_awaited_once()



import asyncio
import logging

from fastapi import Request
from fastapi.responses import StreamingResponse

from bovine.types import ServerSentEvent

logger = logging.getLogger(__name__)


async def async_iterator_for_queue_and_task(
    queue: asyncio.Queue, task, timeout: float = 0.2
):
    try:
        while True:
            try:
                async with asyncio.timeout(timeout):
                    result = await queue.get()
                    if result is None:
                        raise asyncio.CancelledError()
            except TimeoutError:
                logger.debug("Sending heartbeat")
                result = ": <3"
            yield result
    except asyncio.CancelledError:
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass
    except Exception as e:
        logger.exception(e)


def ServerSentEventFromAsyncIterator(
    request: Request, async_iterator
) -> StreamingResponse:
    async def event_stream():
        async for message in async_iterator:
            if await request.is_disconnected():
                return
            data = ServerSentEvent(data=message).encode()
            yield data

    return StreamingResponse(event_stream(), media_type="text/event-stream")


def ServerSentEventFromQueueAndTask(
    request: Request, queue: asyncio.Queue, task: asyncio.Task, timeout: float = 50
):
    """The task is supposed to add strings to queue via
    `await queue.put(some_string)`.

    ```python
    from fastapi import Request

    @fastapi.get("/stream")
    def sample_stream(request: Request):
        queue = asyncio.Queue()
        task = asyncio.create_task(method_to_add_elements_to_queue(queue))

        return ServerSentEventFromQueueAndTask(request, queue, task)
    ```
    """

    async_iterator = async_iterator_for_queue_and_task(queue, task, timeout=timeout)
    return ServerSentEventFromAsyncIterator(request, async_iterator)



from contextlib import asynccontextmanager

from cattle_grid.dependencies import SqlAsyncEngine


def skip_method_information(routing_key):
    """Method information is not created for subscribtions to incoming.* or outgoing.*

    as only cattle_grid should publish to these
    routing keys

    ```pycon
    >>> skip_method_information("incoming.test")
    True

    >>> skip_method_information("outgoing.test")
    True

    >>> skip_method_information("test")
    False

    ```
    """

    if routing_key.startswith("incoming.") or routing_key.startswith("outgoing."):
        return True

    return False


def lifespan_for_sql_alchemy_base_class(base_class):
    """Creates a lifespan that creates the sql alchemy models for the base_class.

    Usage

    ```python
    extension = Extension(
        name="my sample extension",
        module=__name__,
        lifespan=lifespan_for_sql_alchemy_base_class(Base),
    )
    ```
    """

    @asynccontextmanager
    async def lifespan(engine: SqlAsyncEngine):
        async with engine.begin() as conn:
            await conn.run_sync(base_class.metadata.create_all)

        yield

    return lifespan



from dynaconf import Dynaconf

from typing import List

from .validators import all_validators


def get_settings(
    filenames: List[str] = [
        "cattle_grid.toml",
        "cattle_grid_auth.toml",
        "cattle_grid_block_list.toml",
        "config/*.toml",
    ],
) -> Dynaconf:
    return Dynaconf(
        settings_files=filenames,
        envvar_prefix="CATTLE_GRID",
        validators=all_validators,
    )



from faststream.rabbit import RabbitBroker
from faststream import ExceptionMiddleware, Context

import logging

from cattle_grid.model.account import ErrorMessage
from cattle_grid.dependencies import CorrelationId, AccountExchange

from .annotations import AccountName, RoutingKey

logger = logging.getLogger(__name__)

exception_middleware = ExceptionMiddleware()


@exception_middleware.add_handler(Exception)
async def exception_handler(
    exception: Exception,
    name: AccountName,
    routing_key: RoutingKey,
    correlation_id: CorrelationId,
    account_exchange: AccountExchange,
    broker: RabbitBroker = Context(),
):
    logger.error("Processing error occurred for %s", name)
    logger.exception(exception)

    await broker.publish(
        ErrorMessage(message=str(exception), routing_key=routing_key),
        routing_key=f"error.{name}",
        exchange=account_exchange,
        correlation_id=correlation_id,
    )



import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from datetime import datetime
from cattle_grid.testing.fixtures import *  # noqa


from cattle_grid.database.activity_pub_actor import StoredActivity, Follower

from .router_object import ap_router_object


@pytest.fixture
def test_client():
    app = FastAPI()
    app.include_router(ap_router_object)
    yield TestClient(app)


@pytest.fixture
def short_id():
    return "some_id"


@pytest.fixture
def activity_id(short_id):
    return f"http://localhost/object/{short_id}"


@pytest.fixture
async def activity(sql_session_for_test, actor_for_test, short_id, activity_id):  # noqa
    activity_dict = {
        "id": activity_id,
        "actor": actor_for_test.actor_id,
        "type": "Activity",
        "to": ["http://remote.test/actor"],
    }

    sql_session_for_test.add(
        StoredActivity(
            id=short_id,
            actor=actor_for_test,
            data=activity_dict,
            published=datetime.now(),
        )
    )
    await sql_session_for_test.commit()

    return activity_dict


async def test_object_not_found(test_client):
    response = test_client.get(
        "/object/some_id",
        headers={
            "x-cattle-grid-requester": "remote",
            "x-ap-location": "http://localhost/object/some_id",
        },
    )

    assert response.status_code == 404


@pytest.mark.parametrize("requester", ["local", "remote"])
async def test_object_found(
    test_client,
    actor_for_test,
    activity_id,
    activity,
    requester,
):
    response = test_client.get(
        activity_id,
        headers={
            "x-cattle-grid-requester": (
                "http://remote.test/actor"
                if requester == "remote"
                else actor_for_test.actor_id
            ),
            "x-ap-location": activity_id,
        },
    )

    assert response.status_code == 200
    assert response.json() == activity
    assert response.headers["content-type"] == "application/activity+json"


async def test_object_found_follower(
    sql_session_for_test,
    test_client,
    actor_for_test,
    activity_id,
    short_id,
):
    remote_uri = "http://remote.test/actor"

    activity_dict = {
        "id": activity_id,
        "actor": actor_for_test.actor_id,
        "type": "Activity",
        "to": [actor_for_test.followers_uri],
    }
    sql_session_for_test.add_all(
        [
            StoredActivity(
                id=short_id,
                actor=actor_for_test,
                data=activity_dict,
                published=datetime.now(),
            ),
            Follower(
                actor=actor_for_test, follower=remote_uri, accepted=True, request=""
            ),
        ]
    )
    await sql_session_for_test.commit()

    response = test_client.get(
        activity_id,
        headers={
            "x-cattle-grid-requester": remote_uri,
            "x-ap-location": activity_id,
        },
    )

    assert response.status_code == 200
    assert response.json() == activity_dict
    assert response.headers["content-type"] == "application/activity+json"


async def test_object_found_but_unauthorized_no_requester(
    test_client,
    actor_for_test,
    activity_id,
    activity,
):
    response = test_client.get(
        activity_id,
        headers={
            "x-ap-location": activity_id,
        },
    )

    assert response.status_code == 401


async def test_object_found_but_unauthorized(
    test_client,
    actor_for_test,
    activity_id,
    activity,
):
    response = test_client.get(
        activity_id,
        headers={
            "x-cattle-grid-requester": "http://other.test",
            "x-ap-location": activity_id,
        },
    )

    assert response.status_code == 401


async def test_object_wrong_domain(
    test_client,
    actor_for_test,
    activity_id,
    activity,
):
    response = test_client.get(
        activity_id,
        headers={
            "x-cattle-grid-requester": actor_for_test.actor_id,
            "x-ap-location": activity_id.replace("localhost", "otherhost"),
        },
    )

    assert response.status_code == 404



import pytest


from .recipients import extension as recipients_extension
from .webfinger_lookup import extension as webfinger_lookup_extension
from .cache import extension as cache_extension
from .simple_storage import extension as simple_storage_extension


@pytest.mark.parametrize(
    "extension",
    [
        recipients_extension,
        webfinger_lookup_extension,
        cache_extension,
        simple_storage_extension,
    ],
)
def test_factory(extension):
    extension.configure({})



import json

from behave import when, then


@when('"{alice}" sends the trigger action "{method}" with content')  # type: ignore
async def alice_triggers(context, alice, method):
    text_content = context.text
    text_content = text_content.replace(
        f"__{alice}_ID__", context.actors[alice].get("id")
    )

    print()
    print(text_content)
    print()

    await context.connections[alice].trigger(method, json.loads(text_content))


@then('"{alice}" receives an error')  # type: ignore
async def alice_receives_an_error(context, alice):
    error = await context.connections[alice].next_error()

    assert error



from fastapi import FastAPI, APIRouter
from fastapi.testclient import TestClient
from fastapi.responses import HTMLResponse

from . import Extension
from .load import add_routes_to_api


def test_api_router():
    ext = Extension("test", api_prefix="/test", module=__name__)
    result = {"extension": "yes"}

    @ext.get("/")
    async def root_path():
        return result

    ext.configure({})

    app = FastAPI()
    add_routes_to_api(app, [ext])

    test_client = TestClient(app)

    response = test_client.get("/test")

    assert response.status_code == 200

    data = response.json()

    assert data == result


def test_api_router_set_response_class():
    ext = Extension("test", api_prefix="/test", module=__name__)

    @ext.get("/", response_class=HTMLResponse)
    async def root_path():
        return """<html><body>Hi</body></html>"""

    app = FastAPI()
    add_routes_to_api(app, [ext])

    test_client = TestClient(app)

    response = test_client.get("/test")

    assert response.status_code == 200
    assert response.headers["content-type"].startswith("text/html")
    assert response.text == """<html><body>Hi</body></html>"""


def test_api_router_include_router():
    ext = Extension("test", api_prefix="/test", module=__name__)
    result = {"extension": "yes"}

    router = APIRouter()

    @router.get("/")
    async def root_path():
        return result

    ext.include_router(router)

    ext.configure({})

    app = FastAPI()
    add_routes_to_api(app, [ext])

    test_client = TestClient(app)

    response = test_client.get("/test")

    assert response.status_code == 200

    data = response.json()

    assert data == result



import logging
from behave import given, when

from cattle_grid.config import load_settings
from cattle_grid.database import database_session
from cattle_grid.manage.actor import ActorManager
from cattle_grid.model.exchange import UpdateActorMessage

from cattle_grid.testing.features import publish_as, fetch_request

logger = logging.getLogger(__name__)


@given('A new user called "{username}" on "{hostname}"')  # pyright: ignore[reportCallIssue]
def new_user_on_server(context, username, hostname):
    context.execute_steps(
        f"""
        Given An account called "{username}"
        Given "{username}" created an actor on "{hostname}" called "{username}"
        """
    )


@given('A new user called "{username}"')  # pyright: ignore[reportCallIssue]
def new_user(context, username):
    """Creates a new user

    Usage example:

    ```gherkin
    Given A new user called "Alice"
    ```
    """
    hostname = {"alice": "abel", "bob": "banach", "Bob": "banach"}.get(username, "abel")

    context.execute_steps(
        f"""
        Given A new user called "{username}" on "{hostname}"
        """
    )


@when('"{alice}" updates her profile')  # pyright: ignore[reportCallIssue]
async def update_profile(context, alice):
    """
    ```gherkin
    When "Alice" updates her profile
    ```
    """

    for connection in context.connections.values():
        await connection.clear_incoming()

    alice_id = context.actors[alice].get("id")

    msg = UpdateActorMessage(
        actor=alice_id, profile={"summary": "I love cows"}
    ).model_dump()

    await publish_as(context, alice, "update_actor", msg)


@when('"{alice}" fetches her profile')  # pyright: ignore[reportCallIssue]
async def fetch_profile(context, alice):
    """
    ```gherkin
    When "Alice" fetches her profile
    ```

    The profile is stored in `context.profile`
    """

    alice_id = context.actors[alice].get("id")

    context.profile = await fetch_request(context, alice, alice_id)

    assert isinstance(context.profile, dict)


@when('"{alice}" deletes herself')  # pyright: ignore[reportCallIssue]
@when('"{alice}" deletes himself')  # pyright: ignore[reportCallIssue]
async def actor_deletes_themselves(context, alice):
    """
    ```gherkin
    When "Alice" deletes herself
    When "Bob" deletes himself
    ```
    """
    alice_id = context.actors[alice].get("id")

    await publish_as(
        context,
        alice,
        "delete_actor",
        {
            "actor": alice_id,
        },
    )


@given('"{alice}" is in the "{group_name}" group')  # pyright: ignore[reportCallIssue]
async def in_group(context, alice, group_name):
    """
    ```gherkin
    And "Alice" is in the "html" group
    ```
    """
    alice_id = context.actors[alice].get("id")
    config = load_settings()

    async with database_session(db_uri=config.get("db_uri")) as session:  # type: ignore
        manager = ActorManager(session=session, actor_id=alice_id)

        await manager.add_to_group(group_name)



from dataclasses import dataclass
from enum import StrEnum, auto
from typing import List
import http_sf


class ContentType(StrEnum):
    """The content type of the response"""

    activity_pub = auto()
    html = auto()
    other = auto()


@dataclass
class AcceptEntry:
    content_type: str
    profile: str | None = None
    quality: float = 1.0

    @staticmethod
    def from_header(piece):
        return AcceptEntry(
            content_type=str(piece[0]),
            profile=piece[1].get("profile", None),
            quality=float(piece[1].get("q", 1.0)),
        )

    def to_content_type(self) -> ContentType:
        if self.content_type == "text/html":
            return ContentType.html
        elif self.content_type == "application/activity+json":
            return ContentType.activity_pub
        elif (
            self.content_type == "application/ld+json"
            and self.profile == "https://www.w3.org/ns/activitystreams"
        ):
            return ContentType.activity_pub
        else:
            return ContentType.other


def parse_accept_header(header):
    """
    ```pycon
    >>> header = 'application/activity+json, application/ld+json; profile="https://www.w3.org/ns/activitystreams", text/html;q=0.1'
    >>> parse_accept_header(header)
    [AcceptEntry(content_type='application/activity+json', profile=None, quality=1.0),
        AcceptEntry(content_type='application/ld+json', profile='https://www.w3.org/ns/activitystreams', quality=1.0),
        AcceptEntry(content_type='text/html', profile=None, quality=0.1)]

    ```
    """

    parsed = http_sf.parse(header.encode(), tltype="list")

    return [AcceptEntry.from_header(piece) for piece in parsed]  # type:ignore


def should_serve(header: str | None) -> List[ContentType]:
    """
    Determines what content to serve

    ```python
    >>> should_serve("application/activity+json")
    [<ContentType.activity_pub: 'activity_pub'>]

    >>> should_serve("text/html")
    [<ContentType.html: 'html'>]

    ```
    """
    if header is None:
        return [ContentType.other]

    parsed = sorted(parse_accept_header(header), key=lambda x: x.quality, reverse=True)

    return [x.to_content_type() for x in parsed if x.to_content_type() is not None]



import pytest
from sqlalchemy import select
from cattle_grid.extensions.load import build_transformer
from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.extensions.util import lifespan_for_sql_alchemy_base_class

from muck_out.extension import extension as muck_out


from . import extension
from .models import Base, ContextInformation


@pytest.fixture(autouse=True)
async def create_tables(sql_engine_for_tests):
    lifespan = lifespan_for_sql_alchemy_base_class(Base)
    async with lifespan(sql_engine_for_tests):
        yield


@pytest.fixture
def transformer():
    return build_transformer([muck_out, extension])


async def test_transform_accept(transformer):
    result = await transformer(
        {
            "raw": {
                "@context": "https://www.w3.org/ns/activitystreams",
                "id": "http://remote.test/obj_id",
                "type": "Accept",
                "actor": "http://remote.test/actor",
                "object": "http://local.test/follow_id",
                "to": "http://local.test/actor",
            }
        }
    )

    assert result.get("context") == {}


def activity_for_test(object_id: str, in_reply_to: str | None = None):
    obj = {
        "attributedTo": "http://actor.example",
        "content": "<i>italic</i>",
        "id": object_id,
        "published": "2025-09-09T09:32:39Z",
        "to": [
            "https://www.w3.org/ns/activitystreams#Public",
            "http://remote.example/",
        ],
        "type": "Note",
    }

    if in_reply_to:
        obj["inReplyTo"] = in_reply_to

    return {
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            {"Hashtag": "as:Hashtag", "sensitive": "as:sensitive"},
        ],
        "actor": "http://actor.example",
        "id": object_id + "/activity",
        "object": obj,
        "published": "2025-09-09T09:32:39Z",
        "to": [
            "https://www.w3.org/ns/activitystreams#Public",
            "http://remote.example/",
        ],
        "type": "Create",
    }


async def test_transform_create_object(transformer, sql_session):
    object_id = "http://actor.example/object/x36yl8P2Rx8"
    activity = activity_for_test(object_id)

    result = await transformer({"raw": activity})

    context_info = await sql_session.scalar(
        select(ContextInformation).where(ContextInformation.object_id == object_id)
    )

    assert context_info

    assert result["context"]["id"] == str(context_info.context_id)


async def test_transform_create_object_idempotent(transformer):
    object_id = "http://actor.example/object/x36yl8P2Rx8"
    activity = activity_for_test(object_id)

    one = await transformer({"raw": activity})
    two = await transformer({"raw": activity})

    assert one == two


async def test_transform_object_and_reply_same_context(transformer):
    object_id = "http://actor.example/object/x36yl8P2Rx8"
    reply_id = "http://remote.example/object/d7s6d78dsaf"
    activity = activity_for_test(object_id)
    reply = activity_for_test(reply_id, in_reply_to=object_id)

    one = await transformer({"raw": activity})
    two = await transformer({"raw": reply})

    assert one.get("context") == two.get("context")



from fastapi import APIRouter, HTTPException
from sqlalchemy.ext.asyncio import AsyncSession
from cattle_grid.model import FetchMessage
from cattle_grid.dependencies.fastapi import (
    Transformer,
    Broker,
    InternalExchange,
    ActivityExchange,
    SqlSession,
)

from .dependencies import CurrentAccount
from .responses import (
    LookupRequest,
)
from .responses import PerformRequest

from cattle_grid.database.account import Account, ActorForAccount


actor_router = APIRouter(prefix="/actor", tags=["actor"])


async def actor_from_account(
    session: AsyncSession, account: Account, actor_id: str
) -> ActorForAccount | None:
    await session.refresh(account, attribute_names=["actors"])

    for actor in account.actors:
        if actor.actor == actor_id:
            return actor
    return None


@actor_router.post("/lookup", response_model_exclude_none=True, operation_id="lookup")
async def lookup(
    body: LookupRequest,
    account: CurrentAccount,
    broker: Broker,
    exchange: InternalExchange,
    transformer: Transformer,
    session: SqlSession,
) -> dict:
    """Looks up the resource given by `uri` as the actor with
    actor id `actor_id`. Here looking up the actor means that
    the request is signed using a private key belonging to that actor."""

    actor = await actor_from_account(session, account, body.actor_id)
    if actor is None:
        raise HTTPException(400)

    msg = FetchMessage(actor=actor.actor, uri=body.uri)

    result = await broker.publish(
        msg, routing_key="fetch_object", exchange=exchange, rpc=True
    )

    if result is None:
        return {"raw": {}}

    return await transformer({"raw": result})


@actor_router.post("/trigger/{method}", status_code=202, operation_id="trigger")
async def trigger_action(
    method: str,
    body: PerformRequest,
    account: CurrentAccount,
    broker: Broker,
    exchange: ActivityExchange,
    session: SqlSession,
):
    """This method allows one to trigger asynchronous activities
    through a synchronous request. The basic result is that
    the data is posted to the ActivityExchange with the
    routing_key specified.

    """

    actor = await actor_from_account(session, account, body.actor)

    if actor is None:
        raise HTTPException(400)

    await broker.publish(
        body.model_dump(),
        routing_key=method,
        exchange=exchange,
    )



from typing import Callable, Awaitable

from pydantic import BaseModel, Field


class Lookup(BaseModel):
    """
    Lookup of something from the Fediverse
    """

    uri: str = Field(
        examples=["http://actor.example", "acct:user@actor.example"],
        description="""The uri being looked up""",
    )

    actor: str = Field(
        examples=["http://abel.example/actor"],
        description="""The actor performing the lookup""",
    )

    result: dict | None = Field(
        default=None,
        examples=[{"id": "http://actor.example", "type": "Person", "name": "Jane Doe"}],
        description="""The result of the lookup, None if no result yet,
    the result will be returned once the lookup is finished""",
    )


LookupMethod = Callable[[Lookup], Awaitable[Lookup]]
"""Alias for the Lookup Method"""



import pytest
import asyncio
from unittest.mock import AsyncMock
from bovine.activitystreams import Actor

from sqlalchemy.ext.asyncio import async_sessionmaker

from cattle_grid.testing.fixtures import *  # noqa

from .public_key_cache import PublicKeyCache
from bovine.testing import public_key

from cattle_grid.database.auth import Base


async def test_public_key_cache():
    actor = Actor(id="actor_id", public_key=public_key, public_key_name="key").build()
    bovine_actor = AsyncMock()
    bovine_actor.get.return_value = actor

    cache = PublicKeyCache(bovine_actor)

    result = await cache.cryptographic_identifier("some_id")
    assert result
    assert result.controller == "actor_id"

    bovine_actor.get.assert_awaited_once()


async def test_public_key_cache_error():
    bovine_actor = AsyncMock()
    bovine_actor.get.side_effect = Exception("Something went wrong")

    cache = PublicKeyCache(bovine_actor)

    result = await cache.cryptographic_identifier("some_id")
    assert result is None

    bovine_actor.get.assert_awaited_once()


async def test_public_key_cache_gone():
    bovine_actor = AsyncMock()
    bovine_actor.get.return_value = {"type": "Tombstone"}

    cache = PublicKeyCache(bovine_actor)

    result = await cache.cryptographic_identifier("some_id")
    assert result == "gone"

    bovine_actor.get.assert_awaited_once()


@pytest.fixture
async def session_maker(sql_engine_for_tests):
    async with sql_engine_for_tests.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)
    return async_sessionmaker(sql_engine_for_tests)


async def test_cached_public_key(session_maker):
    actor = Actor(id="actor_id", public_key=public_key, public_key_name="key").build()
    bovine_actor = AsyncMock()
    bovine_actor.get.return_value = actor

    cache = PublicKeyCache(bovine_actor, session_maker)

    result = await cache.from_cache("some_id")
    assert result
    assert result.controller == "actor_id"

    result = await cache.from_cache("some_id")
    assert result
    assert result.controller == "actor_id"

    bovine_actor.get.assert_awaited_once()


async def test_cached_public_key_actor_gone(session_maker):
    bovine_actor = AsyncMock()
    bovine_actor.get.return_value = {"type": "Tombstone"}

    cache = PublicKeyCache(bovine_actor, session_maker)

    result = await cache.from_cache("some_id")
    assert result is None

    result = await cache.from_cache("some_id")
    assert result is None

    bovine_actor.get.assert_awaited_once()


async def test_public_key_cache_no_key_result(session_maker):
    actor = Actor(id="actor_id").build()
    bovine_actor = AsyncMock()
    bovine_actor.get.return_value = actor

    cache = PublicKeyCache(bovine_actor, session_maker)

    result = await cache.from_cache("some_id")
    assert result is None

    bovine_actor.get.assert_awaited_once()


async def test_cached_public_key_slow(session_maker):
    actor = Actor(id="actor_id", public_key=public_key, public_key_name="key").build()
    bovine_actor = AsyncMock()

    async def slow_key_return(*args, **kwargs):
        await asyncio.sleep(0.1)
        return actor

    bovine_actor.get.side_effect = slow_key_return

    cache = PublicKeyCache(bovine_actor, session_maker)

    async with asyncio.TaskGroup() as tg:
        task1 = tg.create_task(cache.from_cache("some_id"))
        task2 = tg.create_task(cache.from_cache("some_id"))

    for task in [task1, task2]:
        result = task.result()
        assert result
        assert result.controller == "actor_id"



import click
import os
from cattle_grid.config.auth import new_auth_config, save_auth_config


from .block_cli import add_block_command


@click.group()
def auth_cli(): ...


def add_auth_commands(auth_cli):
    @auth_cli.command()
    @click.argument("actor_id", required=True)
    @click.option(
        "--config_file",
        help="Filename to save the configuration to",
        default="cattle_grid_auth.toml",
    )
    @click.option(
        "--username",
        help="Specify a username for the fetch actor. Otherwise it is automatically generated",
    )
    @click.option(
        "--force",
        help="Used to override the old configuration file",
        is_flag=True,
        default=False,
    )
    def new_config(actor_id, config_file, username, force):
        """Creates a new authorization configuration,
        including generating a public and private key.

        Argument is the actor_id for the fetch actor,
        e.g. http://cattle_grid/fetch_actor
        """

        if os.path.exists(config_file) and not force:
            print("Configuration file already exists! Use --force to overwrite.")
            exit(1)

        config = new_auth_config(actor_id, username=username)

        save_auth_config(config_file, config)


add_auth_commands(auth_cli)
add_block_command(auth_cli)

if __name__ == "__main__":
    auth_cli()



from .examples.simple_storage import extension

from .helper import async_schema_for_extension, openapi_schema_for_extension


def test_async_schema():
    async_schema_for_extension(extension)


def test_openapi_schema():
    result = openapi_schema_for_extension(extension)

    assert isinstance(result, dict)



import logging

from faststream import Context
from faststream.rabbit import RabbitBroker

from bovine.activitystreams.utils import is_public
from sqlalchemy import select

from cattle_grid.dependencies import (
    ActivityExchange,
    CommittingSession,
    CorrelationId,
    SqlSession,
)
from cattle_grid.dependencies.processing import FactoriesForActor

from cattle_grid.extensions import Extension
from cattle_grid.extensions.util import lifespan_for_sql_alchemy_base_class
from cattle_grid.manage.actor import ActorManager
from cattle_grid.model import ActivityMessage
from cattle_grid.model.exchange import UpdateActorMessage
from cattle_grid.model.exchange_update_actor import UpdateActionType, UpdateUrlAction
from cattle_grid.activity_pub.activity import actor_deletes_themselves

from .dependencies import PublishingActor
from .publisher import Publisher

from .config import HtmlDisplayConfiguration
from .models import Base, PublishedObject, PublishingActor as DBPublishingActor
from .router import router
from .types import NameActorMessage


logger = logging.getLogger(__name__)

extension = Extension(
    name="simple html display",
    module=__name__,
    lifespan=lifespan_for_sql_alchemy_base_class(Base),
    config_class=HtmlDisplayConfiguration,
)
extension.rewrite_group_name = "html_display"
extension.rewrite_rules = {"publish_object": "html_display_publish_object"}
extension.include_router(router)


@extension.subscribe("html_display_publish_object")
async def html_publish_object(
    message: ActivityMessage,
    session: CommittingSession,
    actor: PublishingActor,
    config: extension.Config,  # type:ignore
    factories: FactoriesForActor,
    correlation_id: CorrelationId,
    exchange: ActivityExchange,
    broker: RabbitBroker = Context(),
):
    """Publishes an object"""
    obj = message.data

    if not is_public(obj):
        await broker.publish(
            ActivityMessage(actor=message.actor, data=obj),
            routing_key="publish_object",
            exchange=exchange,
            correlation_id=correlation_id,
        )
        return

    if obj.get("id"):
        raise ValueError("Object ID must not be set")

    if obj.get("attributedTo") != message.actor:
        raise ValueError("Actor must match object attributedTo")

    publisher = Publisher(actor, config)
    publisher.update_object(obj)

    session.add(PublishedObject(id=publisher.uuid, data=obj, actor=actor.actor))

    activity = factories[0].create(obj).build()

    await broker.publish(
        ActivityMessage(actor=message.actor, data=activity),
        routing_key="publish_activity",
        exchange=exchange,
        correlation_id=correlation_id,
    )


@extension.subscribe("html_display_name")
async def name_actor(
    message: NameActorMessage,
    actor: PublishingActor,
    session: CommittingSession,
    exchange: ActivityExchange,
    correlation_id: CorrelationId,
    config: extension.Config,  # type:ignore
    broker: RabbitBroker = Context(),
):
    """Sets the display name of the acotr"""
    if message.actor != actor.actor:
        raise Exception("Actor mismatch")

    actor.name = message.name

    await broker.publish(
        UpdateActorMessage(
            actor=actor.actor,
            actions=[
                UpdateUrlAction(
                    action=UpdateActionType.add_url,
                    url=config.html_url(actor.actor, actor.name),
                    media_type="text/html",
                )
            ],
        ),
        routing_key="update_actor",
        exchange=exchange,
        correlation_id=correlation_id,
    )

    if config.automatically_add_users_to_group:
        manager = ActorManager(actor_id=actor.actor, session=session)
        await manager.add_to_group("html_display")


@extension.subscribe("outgoing.Delete")
async def outgoing_delete(
    message: ActivityMessage,
    session: SqlSession,
):
    activity = message.data.get("raw")
    if not isinstance(activity, dict):
        return
    if not actor_deletes_themselves(activity):
        return

    actor = await session.scalar(
        select(DBPublishingActor).where(DBPublishingActor.actor == message.actor)
    )
    if not actor:
        return

    logger.info("Deleting publishing actor with name %s", actor.name)

    await session.delete(actor)
    await session.flush()



from .recipients import transformer


async def test_transformer():
    data = {"raw": {"type": "Activity", "to": ["as:Public"]}}

    result = await transformer(data)

    assert list(result.keys()) == ["recipients"]

    recipients = result["recipients"]

    assert recipients["recipients"] == ["as:Public"]
    assert recipients["public"] is True



import pytest

from sqlalchemy import select, func
from unittest.mock import AsyncMock, MagicMock
from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.database.activity_pub_actor import Follower, Following
from cattle_grid.model import ActivityMessage

from .incoming import (
    incoming_follow_request,
    incoming_accept_activity,
    incoming_reject_activity,
    incoming_delete_activity,
    incoming_block_activity,
    incoming_undo_activity,
)


@pytest.mark.parametrize(
    "object_creator", [lambda x: x, lambda x: {"type": "Person", "id": x}]
)
async def test_incoming_follow_request_create_follower(
    actor_for_test,
    object_creator,
    sql_session_for_test,
):
    msg = ActivityMessage(
        actor=actor_for_test.actor_id,
        data={
            "id": "http://remote.test/id",
            "type": "Follow",
            "actor": "http://remote.test/actor",
            "object": object_creator(actor_for_test.actor_id),
        },
    )
    mock = AsyncMock()

    await incoming_follow_request(
        msg,
        factories=MagicMock(),
        actor=actor_for_test,
        broker=mock,
        session=sql_session_for_test,
    )

    followers = [x for x in await sql_session_for_test.scalars(select(Follower))]

    assert len(followers) == 1

    item = followers[0]
    assert item.follower == "http://remote.test/actor"
    assert not item.accepted

    mock.publish.assert_not_awaited()


async def test_incoming_follow_request_invalid(actor_for_test, sql_session_for_test):
    msg = ActivityMessage(
        actor=actor_for_test.actor_id,
        data={
            "id": "http://remote.test/id",
            "type": "Follow",
            "actor": "http://remote.test/actor",
        },
    )
    await incoming_follow_request(
        msg,
        actor_for_test,
        factories=MagicMock(),
        broker=AsyncMock(),
        session=sql_session_for_test,
    )

    assert 0 == await sql_session_for_test.scalar(func.count(Follower.id))


async def test_incoming_follow_request_auto_follow(
    actor_for_test, sql_session_for_test
):
    actor_for_test.automatically_accept_followers = True
    await sql_session_for_test.merge(actor_for_test)

    msg = ActivityMessage(
        actor=actor_for_test.actor_id,
        data={
            "id": "http://remote.test/id",
            "type": "Follow",
            "actor": "http://remote.test/actor",
            "object": actor_for_test.actor_id,
        },
    )
    mock = AsyncMock()

    factories = [MagicMock(), MagicMock()]
    factories[0].accept.return_value.build.return_value = {"id": "accept:1234"}

    await incoming_follow_request(
        msg,
        factories=factories,  # type: ignore
        actor=actor_for_test,
        broker=mock,
        session=sql_session_for_test,
    )

    mock.publish.assert_awaited_once()


async def test_incoming_accept_activity(sql_session_for_test, actor_for_test):
    follow_id = "follow:1234"
    remote = "http://remote.test/actor"

    following = Following(
        actor=actor_for_test, following=remote, request=follow_id, accepted=False
    )
    sql_session_for_test.add(following)
    await sql_session_for_test.commit()

    msg = ActivityMessage(
        actor=actor_for_test.actor_id,
        data={
            "id": "http://remote.test/id",
            "type": "Accept",
            "actor": remote,
            "object": follow_id,
        },
    )
    await incoming_accept_activity(
        msg, actor_for_test, session=sql_session_for_test, broker=AsyncMock()
    )
    await sql_session_for_test.commit()

    await sql_session_for_test.refresh(following)

    assert following.accepted


async def test_incoming_accept_not_found(sql_session_for_test, actor_for_test):
    remote = "http://remote.test/actor"

    msg = ActivityMessage(
        actor=actor_for_test.actor_id,
        data={
            "id": "http://remote.test/id",
            "type": "Accept",
            "actor": remote,
            "object": "do_not_exist",
        },
    )
    await incoming_accept_activity(
        msg, actor_for_test, session=sql_session_for_test, broker=AsyncMock()
    )


async def test_incoming_reject(sql_session_for_test, actor_for_test):
    broker = AsyncMock()

    sql_session_for_test.add(
        Following(
            actor=actor_for_test,
            following="http://remote.test/",
            accepted=True,
            request="http://actor.test/follow",
        )
    )
    await sql_session_for_test.commit()

    await incoming_reject_activity(
        ActivityMessage(
            actor=actor_for_test.actor_id,
            data={
                "id": "http://actor.test/follow/reject",
                "type": "Reject",
                "actor": "http://remote.test",
                "object": "http://actor.test/follow",
            },
        ),
        actor_for_test,
        broker=broker,
        session=sql_session_for_test,
    )

    following_count = await sql_session_for_test.scalar(func.count(Following.id))

    assert following_count == 0


async def test_incoming_delete(sql_session_for_test, actor_for_test):
    sql_session_for_test.add_all(
        [
            Following(
                actor=actor_for_test,
                following="http://remote.test/",
                accepted=True,
                request="http://actor.test/follow",
            ),
            Follower(
                actor=actor_for_test,
                follower="http://remote.test/",
                accepted=True,
                request="http://actor.test/other",
            ),
        ]
    )

    await incoming_delete_activity(
        ActivityMessage(
            actor=actor_for_test.actor_id,
            data={
                "id": "http://actor.test/follow/delete",
                "type": "Delete",
                "actor": "http://remote.test/",
                "object": "http://remote.test/",
            },
        ),
        session=sql_session_for_test,
    )

    follower_count = await sql_session_for_test.scalar(func.count(Follower.id))
    following_count = await sql_session_for_test.scalar(func.count(Following.id))

    assert follower_count == 0
    assert following_count == 0


async def test_incoming_block(sql_session_for_test, actor_for_test):
    broker = AsyncMock()

    sql_session_for_test.add(
        Following(
            actor=actor_for_test,
            following="http://remote.test/",
            accepted=True,
            request="http://actor.test/follow",
        )
    )
    await sql_session_for_test.commit()

    await incoming_block_activity(
        ActivityMessage(
            actor=actor_for_test.actor_id,
            data={
                "id": "http://actor.test/follow/reject",
                "type": "Block",
                "actor": "http://remote.test/",
                "object": actor_for_test.actor_id,
            },
        ),
        broker=broker,
        actor=actor_for_test,
        session=sql_session_for_test,
    )

    following_count = await sql_session_for_test.scalar(func.count(Following.id))

    assert following_count == 0


@pytest.mark.parametrize(
    "object_builder",
    [
        lambda x: x,
        lambda x: {"type": "Follow", "id": x, "actor": "http://remote.test/"},
    ],
)
async def test_incoming_undo_follow(
    sql_session_for_test, actor_for_test, object_builder
):
    broker = AsyncMock()
    follow_request_id = "http://actor.test/follow"

    sql_session_for_test.add(
        Follower(
            actor=actor_for_test,
            follower="http://remote.test/",
            accepted=True,
            request=follow_request_id,
        )
    )
    await sql_session_for_test.commit()

    await incoming_undo_activity(
        ActivityMessage(
            actor=actor_for_test.actor_id,
            data={
                "id": "http://actor.test/follow/undo",
                "type": "Undo",
                "actor": "http://remote.test/",
                "object": object_builder(follow_request_id),
            },
        ),
        session=sql_session_for_test,
        broker=broker,
        actor=actor_for_test,
    )

    follower_count = await sql_session_for_test.scalar(func.count(Follower.id))

    assert follower_count == 0



import json
from behave import when, then
from almabtrieb.stream import StreamNoNewItemException
from bovine.activitystreams import factories_for_actor_object

from cattle_grid.testing.features import (
    publish_as,
    id_generator_for_actor,
    fetch_request,
    send_message_as_actor,
)


@when('"{actor}" sends "{target}" a message saying "{text}"')  # type: ignore
async def send_message(context, actor, target, text):
    """Used to send a message. The message has the format (with a lot of stuff omitted)

    ```json
    {
        "type": "Create",
        "object": {
            "type": "Note",
            "content": text,
            "to": [actor_id_of_target]
        }
    }
    ```

    This step can be used as

    ```gherkin
    When "alice" sends "bob" a message saying "You stole my milk!"
    ```
    """
    alice = context.actors[actor]
    activity_factory, object_factory = factories_for_actor_object(
        alice, id_generator=id_generator_for_actor(alice)
    )

    bob_id = context.actors[target].get("id")

    note = object_factory.note(content=text, to={bob_id}).build()
    activity = activity_factory.create(note).build()

    await publish_as(
        context,
        actor,
        "send_message",
        send_message_as_actor(alice, activity),
    )


@then('"{actor}" receives an activity')  # type: ignore
async def receive_activity(context, actor):
    """Ensures that an incoming activity was received
    and stores it in `context.activity`.

    ```gherkin
    Then "bob" receives an activity
    ```
    """

    data = await context.connections[actor].next_incoming()
    assert data.get("event_type") == "incoming"

    context.activity = data["data"]["raw"]

    assert context.activity["@context"]


@then('"{actor}" does not receive an activity')  # type: ignore
async def not_receive_activity(context, actor):
    """Ensures that no incoming activity was received

    ```gherkin
    Then "bob" does not receive an activity
    ```
    """

    try:
        result = await context.connections[actor].next_incoming()

        assert result is None, f"Received activity {json.dumps(result, indent=2)}"
    except StreamNoNewItemException:
        ...


@then('the received activity is of type "{activity_type}"')  # type: ignore
def check_activity_type(context, activity_type):
    """Checks that the received activity from [cattle_grid.testing.features.steps.messaging.receive_activity][]
    is of type `activity_type`.

    ```gherkin
    Then the received activity is of type "Update"
    ```
    """

    received = context.activity
    if "raw" in received:
        received = received["raw"]

    import json

    print(json.dumps(received, indent=2))

    assert received.get("type") == activity_type, (
        f"Activity {received} has the wrong type"
    )


@then('"{actor}" receives a message saying "{text}"')  # type: ignore
async def receive_message(context, actor, text):
    """Used to check if the last message received by actor
    is saying the correct thing.

    ```gherkin
    Then "bob" receives a message saying "Got milk?"
    ```

    The received object is stored in `context.received_object`.
    """

    data = await context.connections[actor].next_incoming()

    assert data.get("event_type") == "incoming"
    activity = data.get("data")

    if "raw" in activity:
        activity = activity["raw"]

    assert activity.get("type") == "Create", f"got {activity}"
    assert activity.get("@context"), f"got {activity}"

    obj = activity.get("object", {})
    assert obj.get("content") == text, f"""got {obj.get("content")}"""

    context.received_object = obj


@then('"{bob}" can lookup this message by id')  # type: ignore
async def check_lookup_message(context, bob):
    obj_id = context.received_object.get("id")

    result = await fetch_request(context, bob, obj_id)

    assert isinstance(result, dict), f"got {result}"
    assert result == context.received_object, result


@when('"{actor}" messages her followers "{text}"')  # type: ignore
async def send_message_followers(context, actor, text):
    """Used to send a message to the followers. The message has the format (with a lot of stuff omitted)

    ```json
    {
        "type": "Create",
        "object": {
            "type": "Note",
            "content": text,
            "to": [followers_collection_of_actor]
        }
    }
    ```

    This step can be used as

    ```gherkin
    When "alice" messages her followers "Got milk?"
    ```
    """
    for connection in context.connections.values():
        await connection.clear_incoming()

    alice = context.actors[actor]

    activity_factory, object_factory = factories_for_actor_object(
        alice, id_generator=id_generator_for_actor(alice)
    )
    note = object_factory.note(content=text).as_followers().build()
    activity = activity_factory.create(note).build()

    await publish_as(
        context,
        actor,
        "send_message",
        send_message_as_actor(alice, activity),
    )



"""Added auth table

Revision ID: 2f838b29357f
Revises:
Create Date: 2025-08-25 07:39:07.425115

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "2f838b29357f"
down_revision: Union[str, Sequence[str], None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "cattle_grid_auth_remote_identity",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("key_id", sa.String(length=512), nullable=False),
        sa.Column("controller", sa.String(length=512), nullable=False),
        sa.Column("public_key", sa.String(length=1024), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("key_id"),
        if_not_exists=True,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("cattle_grid_auth_remote_identity")
    # ### end Alembic commands ###



from pydantic import BaseModel, Field

from cattle_grid.extensions.examples.simple_storage.config import determine_url_start


class HtmlDisplayConfiguration(BaseModel):
    prefix: str = Field(
        default="/html_display/object/",
        description="Path to use before the generated uuid. The protocol and domain will be extracted from the actor id. See [cattle_grid.extensions.examples.simple_storage.config.determine_url_start][].",
    )

    html_prefix: str = Field(
        default="/@", description="Prefix to use for the url pages"
    )

    automatically_add_users_to_group: bool = Field(
        default=False,
        description="Causes an actor that sends a html_display_name message to be added to the html_display group",
    )

    def url_start(self, actor_id):
        return determine_url_start(actor_id, self.prefix)

    def html_url_start(self, actor_id):
        """
        Determines the start of an url

        ```python
        >>> config = HtmlDisplayConfiguration()
        >>> config.html_url_start("http://actor.example/some/id")
        'http://actor.example/@'

        ```
        """
        return determine_url_start(actor_id, self.html_prefix)

    def html_url(self, actor_id, actor_name):
        """
        Determines the start of an url

        ```python
        >>> config = HtmlDisplayConfiguration()
        >>> config.html_url("http://actor.example/some/id", "john")
        'http://actor.example/@john'

        ```
        """
        return self.html_url_start(actor_id) + actor_name



import asyncio
import logging
import aio_pika

from cattle_grid.dependencies.fastapi import Broker
from cattle_grid.model.account import EventType

from cattle_grid.tools import enqueue_to_routing_key_and_connection

logger = logging.getLogger(__name__)


def construct_routing_key(account_name: str, event_type: EventType):
    if event_type == EventType.error:
        return f"error.{account_name}"
    return f"receive.{account_name}.{event_type.value}"


async def enqueue_events(
    asyncio_queue, connection, account_name: str, event_type: EventType
):
    routing_key = construct_routing_key(account_name, event_type)

    await enqueue_to_routing_key_and_connection(
        connection, asyncio_queue, routing_key=routing_key, name=account_name
    )


def get_message_streamer(broker: Broker, timeout: float = 5):
    if broker._connection is None:
        raise RuntimeError("Broker not connected")

    connection: aio_pika.RobustConnection = broker._connection

    def stream_messages(account_name: str, event_type: EventType):
        queue = asyncio.Queue()
        task = asyncio.create_task(
            enqueue_events(queue, connection, account_name, event_type)
        )
        return queue, task

    return stream_messages



import pytest

from fastapi import FastAPI

from cattle_grid.testing.fixtures import *  # noqa

from . import (
    async_api_components,
    async_api_schema_for_component,
    fastapi_for_component,
    fastapi_components,
)


@pytest.mark.parametrize("component", async_api_components)
def test_async_api_schema(component, sql_engine_for_tests):
    schema = async_api_schema_for_component(component)
    assert isinstance(schema, str)


@pytest.mark.parametrize("component", fastapi_components)
def test_fastapi_for_component(component, sql_engine_for_tests):
    schema = fastapi_for_component(component)
    assert isinstance(schema, FastAPI)



import logging
from urllib.parse import urlparse

from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.model.exchange_update_actor import UpdateIdentifierAction
from cattle_grid.database.activity_pub_actor import (
    Actor,
    PublicIdentifier,
    PublicIdentifierStatus,
)
from cattle_grid.account.account import account_for_actor
from cattle_grid.account.permissions import allowed_base_urls
from cattle_grid.dependencies.globals import global_container


def get_base_urls() -> list[str]:
    return global_container.config.get("frontend", {}).get(  # type:ignore
        "base_urls", []
    )


logger = logging.getLogger(__name__)


def is_identifier_for_a_base_url(identifier: str, base_urls: list[str]) -> bool:
    domains = [urlparse(base_url).netloc for base_url in base_urls]

    parsed_identifier = urlparse(identifier)

    if parsed_identifier.scheme == "acct":
        identifier_domain = parsed_identifier.path.split("@")[1]
        return identifier_domain in domains
    elif parsed_identifier.scheme in ["http", "https"]:
        return parsed_identifier.netloc in domains

    return False


def is_identifier_part_of_base_urls(identifier: str, base_urls: list[str]) -> bool:
    if not identifier.startswith("acct:"):
        return False

    identifier_domain = identifier.split("@")[1]

    return any(identifier_domain == urlparse(base_url).netloc for base_url in base_urls)


async def determine_identifier_status(identifier):
    base_urls = get_base_urls()

    if is_identifier_part_of_base_urls(identifier, base_urls):
        return PublicIdentifierStatus.verified

    return PublicIdentifierStatus.unverified


def find_identifier(actor: Actor, to_find: str) -> PublicIdentifier | None:
    for identifier in actor.identifiers:
        if identifier.identifier == to_find:
            return identifier
    return None


def new_primary_preference(actor):
    preferences = [identifier.preference for identifier in actor.identifiers]
    if len(preferences) == 0:
        return 0

    return max(*preferences) + 1


async def base_urls_for_actor(session: AsyncSession, actor: Actor) -> list[str]:
    account = await account_for_actor(session, actor.actor_id)

    if account is None:
        return []

    return await allowed_base_urls(session, account)


async def handle_create_identifier(
    actor: Actor, session: AsyncSession, action: UpdateIdentifierAction
) -> None:
    base_urls = await base_urls_for_actor(session, actor)

    logger.info("Got base urls %s", ", ".join(base_urls))
    identifier = action.identifier
    if not is_identifier_for_a_base_url(identifier, base_urls):
        raise ValueError("Cannot create an identifier for this domain")

    await session.refresh(actor, attribute_names=["identifiers"])

    preference = 0

    if action.primary:
        preference = new_primary_preference(actor)

    logger.info(
        "creating identifier %s for %s",
        action.identifier,
        actor.actor_id,
    )

    session.add(
        PublicIdentifier(
            actor=actor,
            identifier=action.identifier,
            name="through_exchange",
            preference=preference,
            status=PublicIdentifierStatus.owned,
        )
    )
    await session.commit()


async def handle_add_identifier(
    actor: Actor, session: AsyncSession, action: UpdateIdentifierAction
):
    """Adds an identifier to the actor

    FIXME: Currently missing logic to verify identifier"""

    base_urls = await base_urls_for_actor(session, actor)
    identifier = action.identifier
    if is_identifier_for_a_base_url(identifier, base_urls):
        raise ValueError(
            "Cannot add an identifier for a controlled base_url use create instead"
        )
    await session.refresh(actor, attribute_names=["identifiers"])

    preference = 0

    if action.primary:
        preference = new_primary_preference(actor)

    session.add(
        PublicIdentifier(
            actor=actor,
            identifier=identifier,
            name="through_exchange",
            preference=preference,
            status=PublicIdentifierStatus.unverified,
        )
    )
    await session.commit()


async def handle_update_identifier(
    actor: Actor, session: AsyncSession, action: UpdateIdentifierAction
) -> None:
    await session.refresh(actor, attribute_names=["identifiers"])

    public_identifier = find_identifier(actor, action.identifier)
    if public_identifier is None:
        raise ValueError("Identifier not found")

    if action.primary:
        public_identifier.preference = new_primary_preference(actor)

    await session.merge(public_identifier)



from pydantic import BaseModel


class ToSendMessage(BaseModel):
    """Internally used to send a message from actor
    to target with the content data"""

    actor: str
    data: dict
    target: str


class StoreActivityMessage(BaseModel):
    """Stores the activity and then sends it, an id is assigned"""

    actor: str
    data: dict



import logging
from sqlalchemy.orm.attributes import flag_modified

from bovine.activitystreams.utils.property_value import from_key_value

from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.model.exchange_update_actor import UpdatePropertyValueAction

logger = logging.getLogger(__name__)


def find_key_in_attachments(attachments: list, key: str) -> int | None:
    for idx, attachment in enumerate(attachments):
        if not isinstance(attachment, dict):
            continue
        if attachment.get("type") != "PropertyValue":
            continue
        if attachment.get("name") == key:
            return idx

    return None


def handle_update_property_value(
    actor: Actor, action: UpdatePropertyValueAction
) -> None:
    current_attachments = actor.profile.get("attachment", [])

    current_index = find_key_in_attachments(current_attachments, action.key)

    if current_index is None:
        current_attachments.append(from_key_value(action.key, action.value))
    else:
        current_attachments[current_index] = from_key_value(action.key, action.value)

    actor.profile["attachment"] = current_attachments
    flag_modified(actor, "profile")

    logger.info("added property %s=%s to %s", action.key, action.value, actor.actor_id)


def handle_delete_property_value(actor: Actor, action: UpdatePropertyValueAction):
    current_attachments = actor.profile.get("attachment", [])

    current_index = find_key_in_attachments(current_attachments, action.key)

    if current_index is not None:
        del current_attachments[current_index]

    actor.profile["attachment"] = current_attachments
    flag_modified(actor, "profile")



"""activity pub and account tables

Revision ID: bab34046008c
Revises: 35d59ad5716c
Create Date: 2025-08-29 11:55:05.146080

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "bab34046008c"
down_revision: Union[str, Sequence[str], None] = "2f838b29357f"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "account",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=256), nullable=False),
        sa.Column("password_hash", sa.String(length=256), nullable=False),
        sa.Column("meta_information", sa.JSON(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "actor",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("actor_id", sa.String(length=256), nullable=False),
        sa.Column("inbox_uri", sa.String(length=256), nullable=False),
        sa.Column("outbox_uri", sa.String(length=256), nullable=False),
        sa.Column("following_uri", sa.String(length=256), nullable=False),
        sa.Column("followers_uri", sa.String(length=256), nullable=False),
        sa.Column("preferred_username", sa.String(length=256), nullable=True),
        sa.Column("public_key_name", sa.String(length=256), nullable=False),
        sa.Column("public_key", sa.Text(), nullable=False),
        sa.Column("automatically_accept_followers", sa.Boolean(), nullable=False),
        sa.Column("profile", sa.JSON(), nullable=False),
        sa.Column("status", sa.String(length=7), nullable=False),
        sa.Column(
            "created_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("actor_id"),
        sa.UniqueConstraint("followers_uri"),
        sa.UniqueConstraint("following_uri"),
        sa.UniqueConstraint("inbox_uri"),
        sa.UniqueConstraint("outbox_uri"),
    )
    op.create_table(
        "credential",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("actor_id", sa.String(length=256), nullable=False),
        sa.Column("identifier", sa.String(length=256), nullable=False),
        sa.Column("secret", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("identifier"),
    )
    op.create_table(
        "inboxlocation",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("actor", sa.String(length=256), nullable=False),
        sa.Column("inbox", sa.String(length=256), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("actor"),
    )
    op.create_table(
        "actorforaccount",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("account_id", sa.Integer(), nullable=False),
        sa.Column("actor", sa.String(length=256), nullable=False),
        sa.Column("name", sa.String(length=256), nullable=False),
        sa.Column("status", sa.String(length=10), nullable=False),
        sa.ForeignKeyConstraint(
            ["account_id"],
            ["account.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "authenticationtoken",
        sa.Column("token", sa.String(length=65), nullable=False),
        sa.Column("account_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["account_id"],
            ["account.id"],
        ),
        sa.PrimaryKeyConstraint("token"),
    )
    op.create_table(
        "blocking",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("actor_id", sa.Integer(), nullable=False),
        sa.Column("blocking", sa.String(length=256), nullable=False),
        sa.Column("request", sa.String(length=256), nullable=False),
        sa.Column("active", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["actor_id"],
            ["actor.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "follower",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("actor_id", sa.Integer(), nullable=False),
        sa.Column("follower", sa.String(length=256), nullable=False),
        sa.Column("request", sa.String(length=256), nullable=False),
        sa.Column("accepted", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["actor_id"],
            ["actor.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "following",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("actor_id", sa.Integer(), nullable=False),
        sa.Column("following", sa.String(length=256), nullable=False),
        sa.Column("request", sa.String(length=256), nullable=False),
        sa.Column("accepted", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["actor_id"],
            ["actor.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "permission",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("account_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=256), nullable=False),
        sa.ForeignKeyConstraint(
            ["account_id"],
            ["account.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "publicidentifier",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("actor_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=256), nullable=False),
        sa.Column("identifier", sa.String(length=256), nullable=False),
        sa.Column("preference", sa.Integer(), nullable=False),
        sa.Column("status", sa.String(length=10), nullable=False),
        sa.ForeignKeyConstraint(
            ["actor_id"],
            ["actor.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("identifier"),
    )
    op.create_table(
        "storedactivity",
        sa.Column("id", sa.String(length=256), nullable=False),
        sa.Column("actor_id", sa.Integer(), nullable=False),
        sa.Column("data", sa.JSON(), nullable=False),
        sa.Column("published", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["actor_id"],
            ["actor.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "actorgroup",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=256), nullable=False),
        sa.Column("actor_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["actor_id"],
            ["actorforaccount.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("actorgroup")
    op.drop_table("storedactivity")
    op.drop_table("publicidentifier")
    op.drop_table("permission")
    op.drop_table("following")
    op.drop_table("follower")
    op.drop_table("blocking")
    op.drop_table("authenticationtoken")
    op.drop_table("actorforaccount")
    op.drop_table("inboxlocation")
    op.drop_table("credential")
    op.drop_table("actor")
    op.drop_table("account")
    # ### end Alembic commands ###



import pytest

from cattle_grid.testing.fixtures import *  # noqa

from . import list_actors, show_actor


@pytest.mark.parametrize("deleted", [True, False])
async def test_list_actors(sql_session_for_test, deleted):
    await list_actors(sql_session_for_test, deleted)


async def test_show_actor(sql_session_for_test, actor_with_account):
    await show_actor(sql_session_for_test, actor_with_account.actor_id)



import logging

from behave import then

from cattle_grid.testing.features import fetch_request

logger = logging.getLogger(__name__)


@then('The "{collection}" collection of "{bob}" does not include "{alice}"')
async def check_collection(context, alice, bob, collection):
    """Used to check if the followers or following collection
    of the actor `bob` does not contain the actor `alice`.

    ```gherkin
    Then The "followers" collection of "bob" does not include "alice"
    ```
    """
    result = await fetch_request(
        context,
        bob,
        context.actors[bob].get(collection),
    )

    assert result

    actor = context.actors[alice].get("id")

    if "raw" in result:
        result = result["raw"]

    assert result.get("type") == "OrderedCollection"
    assert actor not in result.get("orderedItems", [])


@then('The "{collection}" collection of "{bob}" contains "{alice}"')
async def check_collection_contains(context, alice, bob, collection):
    """Used to check if the followers or following collection
    of the actor `bob` contains the actor `alice`.

    ```gherkin
    Then The "followers" collection of "bob" contains "alice"
    ```
    """
    result = await fetch_request(
        context,
        bob,
        context.actors[bob].get(collection),
    )

    bob_id = context.actors[alice].get("id")

    if "raw" in result:
        result = result["raw"]

    assert result.get("type") == "OrderedCollection"
    assert bob_id in result.get("orderedItems", [])



from fastapi import HTTPException
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.database.activity_pub_actor import Actor, ActorStatus
from cattle_grid.activity_pub.actor.relationship import is_blocked


def validate_actor(actor: Actor | None) -> Actor:
    if actor is None:
        raise HTTPException(404)
    if actor.status == ActorStatus.deleted:
        raise HTTPException(410)

    return actor


async def validate_request(
    session: AsyncSession, actor: Actor | None, requester: str | None
) -> Actor:
    if requester is None:
        raise HTTPException(401)

    actor = validate_actor(actor)

    if await is_blocked(session, actor, requester):
        raise HTTPException(403)

    return actor



import pytest

from .http_util import should_serve, ContentType


@pytest.mark.parametrize(
    "header,expected",
    [
        ("", []),
        ("text/html", [ContentType.html]),
        ("application/activity+json", [ContentType.activity_pub]),
    ],
)
def test_should_serve(header, expected):
    assert should_serve(header) == expected


def test_none_as_header():
    assert should_serve(None) == [ContentType.other]



from cattle_grid.testing.fixtures import *  # noqa

from . import ActorManager


async def test_actor_manager_profile(actor_for_test, sql_session_for_test):
    manager = ActorManager(
        actor_id=actor_for_test.actor_id, session=sql_session_for_test
    )

    result = await manager.profile(sql_session_for_test)

    assert result["id"] == actor_for_test.actor_id
    assert len(result["identifiers"]) > 0



import pytest

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.model import ActivityMessage

from cattle_grid.database.activity_pub_actor import Follower, Following

from .util import update_recipients_for_collections


@pytest.fixture
def message(actor_for_test):  # noqa
    return ActivityMessage(actor=actor_for_test.actor_id, data={})


async def test_update_recipients_for_collection(sql_session_for_test, message):
    recipients = {"http://remote.test"}

    result = await update_recipients_for_collections(
        sql_session_for_test, message, recipients
    )

    assert result == recipients


async def test_update_recipients_for_collection_with_followers(
    message, actor_for_test, sql_session_for_test
):
    recipients = {"http://remote.test", actor_for_test.followers_uri}

    sql_session_for_test.add(
        Follower(
            actor=actor_for_test,
            follower="http://follower.test",
            accepted=True,
            request="none",
        )
    )
    await sql_session_for_test.commit()

    result = await update_recipients_for_collections(
        sql_session_for_test, message, recipients
    )

    assert result == {"http://remote.test", "http://follower.test"}


async def test_update_recipients_for_collection_for_following_followers(
    message, actor_for_test, sql_session_for_test
):
    recipients = {
        "http://remote.test",
        actor_for_test.followers_uri,
        actor_for_test.following_uri,
    }

    sql_session_for_test.add_all(
        [
            Follower(
                actor=actor_for_test,
                follower="http://follower.test",
                accepted=True,
                request="none",
            ),
            Following(
                actor=actor_for_test,
                following="http://following.test",
                accepted=True,
                request="none",
            ),
        ]
    )
    await sql_session_for_test.commit()

    result = await update_recipients_for_collections(
        sql_session_for_test, message, recipients
    )

    assert result == {"http://remote.test", "http://follower.test"}


async def test_update_recipients_for_collection_for_self_delete(
    message, actor_for_test, sql_session_for_test
):
    recipients = {
        "http://remote.test",
        actor_for_test.followers_uri,
        actor_for_test.following_uri,
    }

    sql_session_for_test.add_all(
        [
            Follower(
                actor=actor_for_test,
                follower="http://follower.test",
                accepted=True,
                request="none",
            ),
            Following(
                actor=actor_for_test,
                following="http://following.test",
                accepted=True,
                request="none",
            ),
        ]
    )
    await sql_session_for_test.commit()

    message.data = {
        "type": "Delete",
        "actor": actor_for_test.actor_id,
        "object": actor_for_test.actor_id,
    }

    result = await update_recipients_for_collections(
        sql_session_for_test, message, recipients
    )

    assert result == {
        "http://remote.test",
        "http://follower.test",
        "http://following.test",
    }



import json
import aiohttp
import pytest
import pydantic

from fast_depends import dependency_provider
from faststream.rabbit import RabbitBroker, TestRabbitBroker
from unittest.mock import AsyncMock, MagicMock
from collections import namedtuple

from cattle_grid.dependencies import get_client_session
from cattle_grid.dependencies.globals import global_container
from cattle_grid.testing.fixtures import *  # noqa


from .remote import fetch_object, sending_message, resolve_inbox


@pytest.fixture
async def session():
    session_mock = AsyncMock(aiohttp.ClientSession)

    async def get_mock():
        yield session_mock

    dependency_provider.dependency_overrides[get_client_session] = get_mock

    return session_mock


@pytest.fixture
async def broker():
    br = RabbitBroker("amqp://guest:guest@localhost:5672/")
    br.subscriber("to_send", global_container.internal_exchange)(sending_message)
    br.subscriber("fetch_object", global_container.internal_exchange)(fetch_object)

    async with TestRabbitBroker(br, connect_only=False) as tbr:
        yield tbr


@pytest.mark.parametrize(
    "data",
    [
        {},
        {"actor": "http://alice"},
        {"uri": "http://object"},
    ],
)
async def test_fetch_invalid(broker, data, actor_for_test):
    from cattle_grid.dependencies.globals import global_container

    global_container.lookup = AsyncMock(
        return_value=MagicMock(result=None, uri="http://object")
    )

    with pytest.raises(pydantic.ValidationError):
        await broker.publish(
            data,
            routing_key="fetch_object",
            exchange=global_container.internal_exchange,
            rpc=True,
        )


async def test_fetch(broker, session, actor_for_test):
    expected = {"key": "value"}

    Result = namedtuple("http_result", ["status", "raise_for_status", "text"])

    result = Result(200, MagicMock(), AsyncMock(return_value=json.dumps(expected)))
    session.get = AsyncMock(return_value=result)

    from cattle_grid.dependencies.globals import global_container

    global_container.lookup = AsyncMock(
        return_value=MagicMock(result=None, uri="http://object")
    )

    response = await broker.publish(
        {"actor": actor_for_test.actor_id, "uri": "http://object"},
        routing_key="fetch_object",
        exchange=global_container.internal_exchange,
        rpc=True,
    )

    session.get.assert_awaited_once()

    assert response == expected


async def test_to_send_no_inbox(broker, session, actor_for_test):
    Result = namedtuple("http_result", ["status", "raise_for_status", "text"])

    result = Result(
        200,
        MagicMock(),
        AsyncMock(return_value=json.dumps({})),
    )
    session.get = AsyncMock(return_value=result)
    session.post = AsyncMock(return_value=result)

    await broker.publish(
        {
            "actor": actor_for_test.actor_id,
            "target": "http://bob.test/actor",
            "data": {"type": "TestActivity"},
        },
        routing_key="to_send",
        exchange=global_container.internal_exchange,
    )

    session.get.assert_awaited_once()
    session.post.assert_not_called()


async def test_to_send(broker, session, actor_for_test):
    Result = namedtuple("http_result", ["status", "raise_for_status", "text"])

    result = Result(
        200,
        MagicMock(),
        AsyncMock(return_value=json.dumps({"inbox": "http://bob.test/inbox"})),
    )
    session.get = AsyncMock(return_value=result)
    session.post = AsyncMock(return_value=result)

    await broker.publish(
        {
            "actor": actor_for_test.actor_id,
            "target": "http://bob.test/actor",
            "data": {"type": "TestActivity"},
        },
        routing_key="to_send",
        exchange=global_container.internal_exchange,
    )

    session.get.assert_awaited_once()
    session.post.assert_called_once()


async def test_resolve_inbox_no_inbox(sql_session_for_test):
    actor = AsyncMock()
    actor.get = AsyncMock(return_value={})

    result = await resolve_inbox(sql_session_for_test, actor, "target")
    assert result is None


async def test_resolve_inbox(sql_session_for_test):
    actor = AsyncMock()
    actor.get = AsyncMock(return_value={"inbox": "inbox"})

    result = await resolve_inbox(sql_session_for_test, actor, "target")
    assert result == "inbox"

    actor.get.assert_awaited_once()

    # inbox is cached
    result = await resolve_inbox(sql_session_for_test, actor, "target")
    actor.get.assert_awaited_once()



"""Implements loading the configuration"""

from dynaconf import Dynaconf

from typing import List

from .settings import get_settings


default_filenames = [
    "cattle_grid.toml",
    "cattle_grid_auth.toml",
    "cattle_grid_block_list.toml",
    "config/*.toml",
]


def load_settings(filenames: List[str] | None = None) -> Dynaconf:
    if filenames:
        return get_settings(filenames)
    else:
        return get_settings()



from pydantic import BaseModel, Field


class PublishActivity(BaseModel):
    """Used when publishing an activity"""

    actor: str = Field(
        examples=["http://alice.example"],
        description="The actor performing the activity",
    )

    data: dict = Field(
        examples=[
            {
                "@context": "https://www.w3.org/ns/activitystreams",
                "type": "AnimalSound",
                "actor": "http://alice.example",
                "to": ["http://bob.example"],
                "content": "moo",
            }
        ],
        description="""Activity to publish""",
    )


class PublishObject(BaseModel):
    """Used when publishing an object"""

    actor: str = Field(
        examples=["http://alice.example"],
        description="""The actor performing the activity""",
    )

    data: dict = Field(
        examples=[
            {
                "@context": "https://www.w3.org/ns/activitystreams",
                "type": "Note",
                "attributedTo": "http://alice.example",
                "to": ["http://bob.example"],
                "content": "moo",
            }
        ],
        description="""Object to publish""",
    )



import pytest

from unittest.mock import AsyncMock

from faststream.rabbit import RabbitBroker, TestRabbitBroker, RabbitQueue

from cattle_grid.activity_pub.actor import create_actor
from cattle_grid.testing.fixtures import *  # noqa

from .router import create_router
from cattle_grid.database.account import Account, ActorForAccount
from cattle_grid.model.extension import MethodInformationModel
from cattle_grid.dependencies.globals import global_container


@pytest.fixture
async def subscriber_mock():
    return AsyncMock()


@pytest.fixture
async def receive_subscriber_mock():
    return AsyncMock()


@pytest.fixture
async def test_account(sql_session_for_test):
    account = Account(name="alice", password_hash="password")
    sql_session_for_test.add(account)
    await sql_session_for_test.commit()
    return account


@pytest.fixture
async def test_actor(sql_session_for_test, test_account):
    actor = await create_actor(
        sql_session_for_test, "http://localhost/", preferred_username="alice"
    )
    sql_session_for_test.add(
        ActorForAccount(actor=actor.actor_id, account=test_account)
    )
    await sql_session_for_test.commit()
    return actor


@pytest.fixture
async def test_broker(subscriber_mock, receive_subscriber_mock):
    br = RabbitBroker("amqp://guest:guest@localhost:5672/")
    br.include_router(create_router())

    async def mock(msg):
        await subscriber_mock(msg)

        return {"type": "Person", "data": "blank"}

    br.subscriber("send_message", exchange=global_container.exchange)(mock)
    br.subscriber("fetch_object", exchange=global_container.internal_exchange)(mock)

    @br.subscriber(
        RabbitQueue("queue2", routing_key="receive.*.response.*"),
        exchange=global_container.account_exchange,
    )
    async def receive_mock(msg):
        await receive_subscriber_mock(msg)

    async with TestRabbitBroker(br) as tbr:
        yield tbr


async def test_fetch_nothing_happens(test_broker, subscriber_mock):
    await test_broker.publish(
        {"uri": "http://remote/ap/actor/bob"},
        routing_key="send.alice.request.fetch",
        exchange=global_container.account_exchange,
    )

    subscriber_mock.assert_not_called()


async def test_fetch_requires_actor(sql_session_for_test, test_broker, subscriber_mock):
    account = Account(name="alice", password_hash="password")
    actor = await create_actor(
        sql_session_for_test, "http://localhost/", preferred_username="alice"
    )
    sql_session_for_test.add_all(
        [account, ActorForAccount(actor=actor.actor_id, account=account)]
    )
    await sql_session_for_test.commit()

    fetch_uri = "http://remote/ap/actor/bob"

    await test_broker.publish(
        {
            "uri": fetch_uri,
            "actor": "http://localhost/other",
        },
        routing_key="send.alice.request.fetch",
        exchange=global_container.account_exchange,
    )

    subscriber_mock.assert_not_called()


async def test_fetch(test_broker, subscriber_mock, test_actor):
    fetch_uri = "http://remote/ap/actor/bob"

    await test_broker.publish(
        {"uri": fetch_uri, "actor": test_actor.actor_id},
        routing_key="send.alice.request.fetch",
        exchange=global_container.account_exchange,
    )

    subscriber_mock.assert_called_once()
    args = subscriber_mock.call_args[0][0]

    assert args["uri"] == fetch_uri
    assert args["actor"] == test_actor.actor_id


async def test_getting_info(test_broker, receive_subscriber_mock, test_actor):
    global_container.method_information = [
        MethodInformationModel(routing_key="test", module="test")
    ]

    await test_broker.publish(
        {"action": "info", "data": {}, "actor": ""},
        routing_key="send.alice.request.info",
        exchange=global_container.account_exchange,
    )

    receive_subscriber_mock.assert_called_once()
    args = receive_subscriber_mock.call_args[0][0]

    assert args["actors"] == [{"id": test_actor.actor_id, "name": "NO NAME"}]

    assert len(args["methodInformation"]) > 0



import logging

from typing import Annotated
from faststream import Depends

from bovine import BovineActor

from cattle_grid.dependencies import ClientSession, SqlSession
from cattle_grid.dependencies.processing import actor_id, ProcessingError

from cattle_grid.activity_pub.actor import (
    bovine_actor_for_actor_id,
)

logger = logging.getLogger(__name__)


async def bovine_actor_for_message(
    actor_id: Annotated[str, Depends(actor_id)],
    session: ClientSession,
    sql_session: SqlSession,
) -> BovineActor:
    actor = await bovine_actor_for_actor_id(sql_session, actor_id)
    if actor is None:
        raise ProcessingError("Actor not found")

    await actor.init(session=session)

    return actor


MessageBovineActor = Annotated[BovineActor, Depends(bovine_actor_for_message)]



from collections.abc import Callable, Awaitable

from fast_depends import inject
from cattle_grid.extensions import Extension
from .types import Transformer


def resolvable_transformers(transformers: set[Transformer], available_fields: set[str]):
    """
    Determines the set of resolvable transformers based on the available fields.

    """
    return {
        transformer
        for transformer in transformers
        if transformer.inputs.issubset(available_fields)
    }


def get_transformers(extensions: list[Extension]) -> set[Transformer]:
    result = set()

    for extension in extensions:
        if extension.transformer:
            if not extension.transformer_inputs or not extension.transformer_outputs:
                raise ValueError(
                    "transformer_inputs and transformer_outputs are required when transformer is set"
                )

            result.add(
                Transformer(
                    name=extension.name,
                    transformer=extension.transformer,
                    inputs=set(extension.transformer_inputs),
                    outputs=set(extension.transformer_outputs),
                )
            )

    return result


def transformation_steps(transformers: set[Transformer]) -> list[set[Transformer]]:
    to_process = transformers.copy()
    steps = []

    available_fields = {"raw"}

    while len(to_process) > 0:
        resolvable = resolvable_transformers(to_process, available_fields)

        if len(resolvable) == 0:
            raise Exception("Could not resolve plugins")
        steps.append(resolvable)
        for x in resolvable:
            available_fields = available_fields | x.outputs
        to_process = to_process - resolvable

    return steps


def build_transformer(extensions: list[Extension]) -> Callable[[dict], Awaitable[dict]]:
    """Builds the transformer"""
    transformers = get_transformers(extensions)
    steps = transformation_steps(transformers)

    async def transformer(data: dict):
        for step in steps:
            for plugin in step:
                data.update(await inject(plugin.transformer)(data))

        return data

    return transformer



import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.account.account import create_account

from cattle_grid.dependencies.globals import global_container

from . import router


@pytest.fixture
def test_client():
    app = FastAPI()
    app.include_router(router)

    return TestClient(app)


@pytest.fixture
async def test_account(sql_session_for_test):
    account = await create_account(sql_session_for_test, "name", "pass")

    return account


def test_signin_missing_data(test_client, test_account):
    response = test_client.post("/signin", json={"name": "name"})

    assert response.status_code == 422


def test_signin_wrong_password(test_client, test_account):
    response = test_client.post("/signin", json={"name": "name", "password": "wrong"})

    assert response.status_code == 401


def test_signin(test_client, test_account):
    response = test_client.post("/signin", json={"name": "name", "password": "pass"})

    global_container.method_information = []

    assert response.status_code == 200

    data = response.json()

    assert "token" in data

    response = test_client.get(
        "/account/info", headers={"Authorization": f"Bearer {data['token']}"}
    )

    assert response.status_code == 200


@pytest.mark.parametrize("endpoint", ["/account/info"])
def test_unauthorized_without_signin(endpoint, test_client):
    response = test_client.get(endpoint)
    assert response.status_code == 401



from unittest.mock import AsyncMock
import pytest

from cattle_grid.extensions.testing import with_test_broker_for_extension
from cattle_grid.model import ActivityMessage
from cattle_grid.dependencies.globals import global_container
from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.extensions.util import lifespan_for_sql_alchemy_base_class
from .models import Base
from . import extension


@pytest.fixture(autouse=True)
async def create_tables(sql_engine_for_tests):
    lifespan = lifespan_for_sql_alchemy_base_class(Base)
    async with lifespan(sql_engine_for_tests):
        yield


@pytest.fixture
def mock_publish_activity():
    return AsyncMock()


@pytest.fixture
def mock_publish_object():
    return AsyncMock()


@pytest.fixture
async def test_broker(mock_publish_activity, mock_publish_object):
    extension.configure({})

    async with with_test_broker_for_extension(
        [extension],
        {
            "publish_activity": mock_publish_activity,
            "publish_object": mock_publish_object,
        },
    ) as tbr:
        yield tbr


@pytest.fixture
async def published_object(actor_for_test, test_broker, mock_publish_activity):
    obj = {
        "type": "Note",
        "to": ["as:Public"],
        "content": "I <3 milk!",
        "attributedTo": actor_for_test.actor_id,
    }

    await test_broker.publish(
        ActivityMessage(actor=actor_for_test.actor_id, data=obj),
        routing_key="html_display_publish_object",
        exchange=global_container.exchange,
    )

    mock_publish_activity.assert_awaited_once()

    args = mock_publish_activity.await_args

    activity = args[1]["data"]

    assert activity["type"] == "Create"

    return activity["object"]



from typing import Dict, Any
from pydantic import BaseModel, ConfigDict, Field


class CreateActorMessage(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    baseUrl: str = Field(
        examples=["http://local.example/"],
        description="""
    base url used to create the user on. Can contain a path
    """,
    )
    preferredUsername: str | None = Field(
        None,
        examples=["alice", "bob"],
        description="""
    Add a preferred username. This name will be used in acct:username@domain and supplied to webfinger. Here domain is determine from baseUrl.
    """,
    )
    profile: Dict[str, Any] = Field(
        {},
        examples=[{"summary": "A new actor"}],
        description="""
    New profile object for the actor. The fields.
    """,
    )
    autoFollow: bool = Field(
        False,
        examples=[True],
        description="""
    Enables setting actors to automatically accept follow requests
    """,
    )



from .rabbitmq import enqueue_to_routing_key_and_connection
from .server_sent_events import ServerSentEventFromQueueAndTask

__all__ = ["enqueue_to_routing_key_and_connection", "ServerSentEventFromQueueAndTask"]



import asyncio
import click
from sqlalchemy import select, delete
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.database import run_with_database
from cattle_grid.database.activity_pub_actor import Actor, ActorStatus
from cattle_grid.database.account import ActorForAccount

from cattle_grid.manage import ActorManager


async def list_actors(session: AsyncSession, deleted: bool = False):
    result = await session.scalars(select(Actor))

    if deleted:
        result = [x for x in result if x.status == ActorStatus.deleted]

    count = 0
    for x in result:
        print(x.actor_id)
        count = count + 1

    if count == 0:
        print("No actors")


async def show_actor(session: AsyncSession, actor_id: str):
    manager = ActorManager(session=session, actor_id=actor_id)
    groups = ", ".join(await manager.groups())

    print(f"Actor ID: {actor_id}")
    if groups == "":
        print("Not a member of a group")
    else:
        print(f"Groups: {groups}")
    print()


async def modify_actor(session: AsyncSession, actor_id: str, add_groups: list[str]):
    manager = ActorManager(session=session, actor_id=actor_id)

    for group_name in add_groups:
        await manager.add_to_group(group_name)


async def prune_actors(session: AsyncSession):
    await session.execute(delete(Actor).where(Actor.status == ActorStatus.deleted))
    await session.execute(delete(ActorForAccount).where(ActorForAccount.deleted))


def add_actors_to_cli_as_group(main):
    @main.group()
    def actor():
        """Used to manage actors"""

    add_to_cli(actor)


def add_to_cli(main):
    @main.command("list")  # type: ignore
    @click.option(
        "--deleted", is_flag=True, default=False, help="Only list deleted actors"
    )
    @click.pass_context
    def list_actors_command(ctx, deleted):
        asyncio.run(
            run_with_database(
                ctx.obj["config"], lambda session: list_actors(session, deleted)
            )
        )

    @main.command("prune")  # type: ignore
    @click.pass_context
    def prune_actors_command(ctx):
        asyncio.run(
            run_with_database(ctx.obj["config"], lambda session: prune_actors(session))
        )

    @main.command("show")  # type: ignore
    @click.argument("actor_id")
    @click.pass_context
    def show_actor_command(ctx, actor_id):
        asyncio.run(
            run_with_database(
                ctx.obj["config"], lambda session: show_actor(session, actor_id)
            )
        )

    @main.command("modify")  # type: ignore
    @click.argument("actor_id")
    @click.option("--add_group", multiple=True, default=[])
    @click.pass_context
    def modify_actor_command(ctx, actor_id, add_group):
        """Adds a group to the actor"""
        asyncio.run(
            run_with_database(
                ctx.obj["config"],
                lambda session: modify_actor(session, actor_id, add_group),
            )
        )



from sqlalchemy.orm.attributes import flag_modified

from bovine.activitystreams.utils import as_list

from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.model.exchange_update_actor import UpdateUrlAction


def build_link_for_action(action: UpdateUrlAction):
    result = {"type": "Link", "href": action.url}
    if action.media_type:
        result["mediaType"] = action.media_type
    if action.rel:
        result["rel"] = action.rel

    return result


def url_matches(item: str | dict, url: str) -> bool:
    """Checks if the item represents the url
    ```
    >>> url_matches("http://remote.test", "http://remote.test")
    True

    >>> url_matches({"href": "http://remote.test"}, "http://remote.test")
    True

    ```
    """
    if isinstance(item, dict):
        return item.get("href") == url

    return item == url


def add_url(actor: Actor, action: UpdateUrlAction):
    urls = as_list(actor.profile.get("url", []))
    urls = [url for url in urls if not url_matches(url, action.url)]

    urls.append(build_link_for_action(action))
    actor.profile["url"] = urls

    flag_modified(actor, "profile")


def remove_url(actor: Actor, action: UpdateUrlAction):
    urls = as_list(actor.profile.get("url", []))

    actor.profile["url"] = [url for url in urls if not url_matches(url, action.url)]
    flag_modified(actor, "profile")



import logging

from typing import Annotated

from fastapi import HTTPException, Request, Response, APIRouter, Header
from fastapi.responses import PlainTextResponse
from starlette.datastructures import MutableHeaders

from pydantic import BaseModel, Field

from bovine.utils import webfinger_response
from bovine.types.jrd import JrdData

from cattle_grid.activity_pub.server.router import ActivityResponse
from .util import check_block
from .http_util import should_serve, ContentType

from .dependencies import (
    AuthConfig as AuthConfigDependency,
    ActorObject,
    SignatureCheckWithCache,
)

logger = logging.getLogger(__name__)


class ReverseProxyHeaders(BaseModel):
    """Headers set by the reverse proxy"""

    x_original_method: str = Field("get", description="""The original used method""")
    x_original_uri: str | None = Field(None, description="""The original request uri""")
    x_original_host: str | None = Field(None, description="""The original used host""")
    x_forwarded_proto: str = Field("http", description="""The protocol being used""")


auth_router = APIRouter(tags=["auth"])
"""The authentication router"""


@auth_router.get("/.well-known/webfinger")
async def webfinger(resource: str, config: AuthConfigDependency) -> JrdData:
    """If resource is the actor corresponding to the actor fetching
    public keys, returns the corresponding Jrd. Otherwise returns
    not found"""
    logger.info(config)
    if resource != config.actor_acct_id:
        raise HTTPException(404)
    return webfinger_response(config.actor_acct_id, config.actor_id)


@auth_router.get(
    "/cattle_grid_actor",
    response_class=ActivityResponse,
)
async def handle_get_actor(actor_object: ActorObject):
    """Returns the actor profile of the
    fetch actor used to retrieve public keys, e.g.

    ```json
    {
        "type": "Service",
        "id": "https://your-domain.example/cattle_grid_actor",
        ...
    }
    ```
    """
    return actor_object


@auth_router.get(
    "/auth",
    responses={
        200: {"description": "Request is valid", "content": {"text/plain": ""}},
        401: {"description": "The signature was invalid"},
        403: {"description": "Request was blocked"},
    },
    response_class=PlainTextResponse,
)
async def verify_signature(
    request: Request,
    response: Response,
    config: AuthConfigDependency,
    signature_checker: SignatureCheckWithCache,
    reverse_proxy_headers: Annotated[ReverseProxyHeaders, Header()],
) -> str:
    """Takes the request and checks signature. If signature check
    fails a 401 is returned. If the domain the public key belongs
    to is blocked, a 403 is returned.

    If the request is valid. The controller corresponding to
    the signature is set in the response header `X-CATTLE-GRID-REQUESTER`.

    The header `X-CATTLE-GRID-SHOULD-SERVE` is set to `html`
    if one should redirect to the HTML resource. It is set to `other` if the resource to serve cannot be determined.
    This is only used for unsigned requests.

    Note: More headers than the ones listed below can be used
    to verify a signature.
    """
    headers = MutableHeaders(request.headers)

    logger.debug(headers)

    logger.info("Got accept header %s", headers.get("accept"))

    servable_content_types = should_serve(headers.get("accept"))

    if "signature" not in headers:
        if ContentType.html in servable_content_types:
            response.headers["x-cattle-grid-should-serve"] = "html"
            return ""
        elif ContentType.other in servable_content_types:
            response.headers["x-cattle-grid-should-serve"] = "other"
            return ""
        elif config.require_signature_for_activity_pub:
            raise HTTPException(401)
        else:
            return ""

    if reverse_proxy_headers.x_original_host:
        headers["Host"] = reverse_proxy_headers.x_original_host

    url = f"{reverse_proxy_headers.x_forwarded_proto}://{reverse_proxy_headers.x_original_host}{reverse_proxy_headers.x_original_uri}"

    logger.debug("Treating request as to url %s", url)

    controller = await signature_checker.validate_signature(
        reverse_proxy_headers.x_original_method.lower(),
        url,
        dict(headers.items()),
        None,
    )

    logger.debug("Got controller %s", controller)

    if controller:
        if check_block(config.domain_blocks, controller):
            logger.info("Blocked a request by %s", controller)
            raise HTTPException(403)
        response.headers["x-cattle-grid-requester"] = controller

        logger.debug("Got requester %s", controller)

        return ""

    logger.info(
        "invalid signature for request to %s => access denied",
        request.headers.get("X-Original-Uri", ""),
    )

    raise HTTPException(401)



import pytest

from sqlalchemy.ext.asyncio import async_sessionmaker

from cattle_grid.account.account import create_account, add_actor_to_account
from cattle_grid.activity_pub.actor import create_actor

from cattle_grid.config.auth import new_auth_config, save_auth_config
from cattle_grid.dependencies.globals import global_container

from cattle_grid.database.activity_pub import Base as APBase


@pytest.fixture(autouse=True)
async def sql_engine_for_tests():
    """Provides the sql engine (as in memory sqlite) for tests"""
    async with global_container.alchemy_database(
        "sqlite+aiosqlite:///:memory:", echo=False
    ) as engine:
        async with engine.begin() as conn:
            await conn.run_sync(APBase.metadata.create_all)

        yield engine


@pytest.fixture()
async def session_maker_for_tests(sql_engine_for_tests):
    yield async_sessionmaker(sql_engine_for_tests, expire_on_commit=False)


@pytest.fixture()
async def sql_session_for_test(session_maker_for_tests):
    async with session_maker_for_tests() as session:
        yield session


@pytest.fixture()
async def sql_session(session_maker_for_tests):
    async with session_maker_for_tests() as session:
        yield session


@pytest.fixture
async def account_for_test(sql_session_for_test):
    """Fixture to create an account"""
    return await create_account(
        sql_session_for_test, "alice", "alice", permissions=["admin"]
    )


@pytest.fixture
async def actor_for_test(sql_session_for_test):
    """Fixture to create an actor"""
    actor = await create_actor(sql_session_for_test, "http://localhost/ap")

    return actor


@pytest.fixture
async def actor_with_account(sql_session_for_test, account_for_test):
    """Fixture to create an actor with an account"""
    actor = await create_actor(sql_session_for_test, "http://localhost/ap")
    await add_actor_to_account(
        sql_session_for_test, account_for_test, actor, name="test_fixture"
    )

    await sql_session_for_test.refresh(actor)

    return actor


@pytest.fixture
def auth_config():
    config = new_auth_config(actor_id="http://localhost/actor_id", username="actor")

    config.domain_blocks = set(["blocked.example"])

    return config


@pytest.fixture
def auth_config_file(tmp_path, auth_config):
    filename = tmp_path / "auth_config.toml"

    save_auth_config(filename, auth_config)

    return filename


@pytest.fixture(autouse=True, scope="session")
def loaded_config():
    """Ensures the configuration variables are loaded"""
    global_container.load_config()



import pytest

from cattle_grid.database.activity_pub_actor import Follower
from cattle_grid.testing.fixtures import *  # noqa

from .relationship import followers_for_actor


@pytest.mark.parametrize("accepted,expected", [(True, 1), (False, 0)])
async def test_followers_for_actor(
    sql_session_for_test, actor_for_test, accepted, expected
):
    follower = Follower(
        actor=actor_for_test,
        follower="http://remote.test/follower",
        request="http://remote.test/request_id",
        accepted=accepted,
    )

    sql_session_for_test.add(follower)
    await sql_session_for_test.commit()

    assert (
        len(await followers_for_actor(sql_session_for_test, actor_for_test)) == expected
    )



from typing import Annotated

from fastapi import Depends

from bovine import BovineActor as BovineActorClass
from bovine.activitystreams import Actor
from bovine.utils import parse_fediverse_handle
from bovine.crypto.signature_checker import SignatureChecker

from cattle_grid.dependencies.fastapi import Config, ClientSession
from cattle_grid.config.auth import AuthConfig as AuthConfigObject, get_auth_config
from cattle_grid.dependencies.globals import global_container

from .util import config_to_bovine_actor
from .public_key_cache import PublicKeyCache


def provide_auth_config(config: Config) -> AuthConfigObject:
    return get_auth_config(config)


AuthConfig = Annotated[AuthConfigObject, Depends(provide_auth_config)]
"""Provides the configuration for the auth module"""


def create_bovine_actor(config: AuthConfig) -> BovineActorClass:
    return config_to_bovine_actor(config)


BovineActor = Annotated[BovineActorClass, Depends(create_bovine_actor)]
"""Returns the bovine actor"""


def create_actor_object(config: AuthConfig) -> dict:
    username, _ = parse_fediverse_handle(config.actor_acct_id.removeprefix("acct:"))
    return Actor(
        id=config.actor_id,
        type="Service",
        public_key=config.public_key,
        preferred_username=username,
        public_key_name="mykey",
    ).build()


ActorObject = Annotated[dict, Depends(create_actor_object)]


async def create_signature_checker(config: AuthConfig, session: ClientSession):
    bovine_actor = config_to_bovine_actor(config)
    await bovine_actor.init(session)
    public_key_cache = PublicKeyCache(
        bovine_actor, global_container.get_session_maker()
    )
    return SignatureChecker(public_key_cache.from_cache, skip_digest_check=True)


SignatureCheckWithCache = Annotated[SignatureChecker, Depends(create_signature_checker)]



import pytest

from unittest.mock import AsyncMock

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.dependencies.globals import get_transformer

from .testing import *  # noqa


@pytest.mark.parametrize(
    "endpoint,data",
    [
        ("/actor/lookup", {"actorId": "bad_actor", "uri": "http://remote.test/"}),
        ("/actor/trigger/method", {"actor": "bad_actor", "data": {}, "method": "none"}),
    ],
)
def test_bad_actor(test_app, test_client, bearer_header, endpoint, data):
    test_app.dependency_overrides[get_transformer] = lambda: AsyncMock(
        side_effect=lambda x: x
    )

    result = test_client.post(
        endpoint,
        json=data,
        headers=bearer_header,
    )

    assert result.status_code == 400


def test_lookup_no_result(test_app, test_client, bearer_header, test_broker):
    test_app.dependency_overrides[get_transformer] = lambda: AsyncMock(
        side_effect=lambda x: x
    )

    result = test_client.post(
        "/account/create",
        json={"baseUrl": "http://abel.test"},
        headers=bearer_header,
    )
    assert result.status_code == 201
    actor_id = result.json()["id"]

    test_broker.publish = AsyncMock(return_value=None)
    result = test_client.post(
        "/actor/lookup",
        json={"actorId": actor_id, "uri": "http://remote.test/"},
        headers=bearer_header,
    )

    assert result.status_code == 200
    assert result.json() == {"raw": {}}


def test_lookup_result(test_app, test_client, bearer_header, test_broker):
    result = test_client.post(
        "/account/create",
        json={"baseUrl": "http://abel.test"},
        headers=bearer_header,
    )
    assert result.status_code == 201
    actor_id = result.json()["id"]

    test_app.dependency_overrides[get_transformer] = lambda: AsyncMock(
        side_effect=lambda x: x
    )

    test_broker.publish = AsyncMock(return_value={"data": 1})

    result = test_client.post(
        "/actor/lookup",
        json={"actorId": actor_id, "uri": "http://remote.test/"},
        headers=bearer_header,
    )

    assert result.status_code == 200
    assert result.json() == {"raw": {"data": 1}}


def test_perform_action(test_app, test_client, bearer_header, test_broker):
    result = test_client.post(
        "/account/create",
        json={"baseUrl": "http://abel.test"},
        headers=bearer_header,
    )
    assert result.status_code == 201
    actor_id = result.json()["id"]

    data = {"actor": actor_id, "data": {"moo": "yes"}}

    result = test_client.post(
        "/actor/trigger/method",
        json=data,
        headers=bearer_header,
    )

    assert result.status_code == 202
    test_broker.publish.assert_awaited_once()

    args = test_broker.publish.call_args.args[0]

    assert args == data



from unittest.mock import AsyncMock, MagicMock
import asyncio
import pytest
from .enqueuer import enqueue_from_inbox


@pytest.fixture
async def mock_broker():
    return AsyncMock()


async def test_enqueue_from_inbox_requires_type(mock_broker):
    await enqueue_from_inbox(mock_broker, MagicMock(), "actor_id", {})

    await asyncio.sleep(0.1)

    mock_broker.publish.assert_not_awaited()


async def test_enqueue_from_inbox(mock_broker):
    await enqueue_from_inbox(
        mock_broker, MagicMock(), "actor_id", {"type": "Test", "actor": "my_actor"}
    )

    await asyncio.sleep(0.1)

    mock_broker.publish.assert_awaited_once()



import os
import json
from click.testing import CliRunner

from .__main__ import main


def test_open_api(tmp_path):
    filename = tmp_path / "openapi.json"
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            "openapi",
            "cattle_grid.extensions.examples.simple_storage",
            "--filename",
            str(filename),
        ],
    )

    assert result.exit_code == 0
    assert os.path.exists(filename)

    with open(filename, "r") as fp:
        schema = json.load(fp)

    assert isinstance(schema, dict)



from faststream import FastStream
from faststream.asyncapi.schema import Schema


from cattle_grid.version import __version__

from cattle_grid.model.account import EventInformation
from .router import create_router


def get_mock_faststream_app() -> FastStream:
    """Creates a mock faststream app for Account processing"""

    from faststream.rabbit import RabbitBroker

    broker = RabbitBroker()
    broker.include_router(create_router())

    broker.publisher(
        "incoming",
        title="receive.NAME.incoming",
        schema=EventInformation,
        description="""Incoming messages from the Fediverse""",
    )

    broker.publisher(
        "outgoing",
        title="receive.NAME.outgoing",
        schema=EventInformation,
        description="""Messages being sent towards the Fediverse""",
    )

    return FastStream(
        broker,
        title="cattle_grid Cattle Drive Implementation",
        version=__version__,
        description="Illustrates how cattle grid processes ActivityPub",
    )


def get_async_api_schema() -> Schema:
    """Returns the async api schema for cattle_grid Account processing"""
    from faststream.asyncapi import get_app_schema

    app = get_mock_faststream_app()

    return get_app_schema(app)



import click
import json
import uvicorn
import logging

from cattle_grid.extensions.load import load_extension

from .helper import (
    async_schema_for_extension,
    openapi_schema_for_extension,
    fastapi_for_extension,
)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def run_app(module, port, host, no_broker):
    extension = load_extension({"module": module})
    app = fastapi_for_extension(extension, include_broker=not no_broker)

    uvicorn.run(app, port=port, host=host)


def build_filename(filename: str | None, api_type: str, name: str) -> str:
    """
    ```
    >>> build_filename("file.txt", "openapi", "some name")
    'file.txt'

    >>> build_filename(None, "openapi", "simple html display")
    './docs/assets/schemas/openapi_simple_html_display.json'

    ```
    """
    if filename:
        return filename

    name = name.replace(" ", "_")
    return f"./docs/assets/schemas/{api_type}_{name}.json"


def add_extension_commands(main):
    @main.command("load")
    @click.argument("module")
    def load_extension_command(module):
        """Loads an extension"""
        load_extension({"module": module})

        # FIXME config parameters

    @main.command("async-api")
    @click.argument("module")
    @click.option("--filename", default=None, help="Filename to write to")
    def async_api(module, filename):
        """Generates the async api schema for the extension"""
        extension = load_extension({"module": module})
        schema = async_schema_for_extension(extension).to_json()
        filename = build_filename(filename, "asyncapi", extension.name)

        with open(filename, "w") as fp:
            fp.write(schema)

        click.echo(f"Wrote async api schema to {filename}")

    @main.command("openapi")
    @click.argument("module")
    @click.option("--filename", default=None, help="Filename to write to")
    def openapi(module, filename):
        """Generates the openapi schema for the extension"""
        extension = load_extension({"module": module})
        schema = openapi_schema_for_extension(extension)
        filename = build_filename(filename, "openapi", extension.name)

        with open(filename, "w") as fp:
            json.dump(schema, fp)

        click.echo(f"Wrote openapi schema to {filename}")

    @main.command("run")
    @click.argument("module")
    @click.option("--host", default="0.0.0.0", help="Host to run on")
    @click.option("--port", default=80, help="Port to run on")
    @click.option("--reload_dir", default=None, help="Reload on changes in directory")
    @click.option(
        "--no_broker",
        is_flag=True,
        default=False,
        help="Set to run without included broker",
    )
    def run_server(module, host, port, reload_dir, no_broker):
        """Runs the extension as an independent server process.
        The configuration is taken from the same files as cattle_grid.
        Thus these must be present."""

        if reload_dir:
            import watchfiles

            logger.info("Watching for changes in %s", reload_dir)

            watchfiles.run_process(
                reload_dir, target=run_app, args=(module, port, host, no_broker)
            )
        else:
            run_app(module, port, host, no_broker)


def add_extension_commands_as_group(main):
    @main.group("extensions")
    def extensions():
        """Commands for managing extensions"""

    add_extension_commands(extensions)



def acct_to_fediverse(acct_uri: str) -> str:
    return acct_uri.replace("acct:", "@")


def gather_property_values(attachments: list[dict] | None) -> list[dict]:
    if not attachments:
        return []
    return [
        {"name": a.get("name"), "value": a.get("value")}
        for a in attachments
        if a.get("type") == "PropertyValue"
    ]


def format_actor_profile(profile: dict) -> dict:
    acct_uris = [x for x in profile.get("identifiers", []) if x.startswith("acct:")]
    return {
        "summary": profile.get("summary"),
        "fediverse_handles": [acct_to_fediverse(x) for x in acct_uris],
        "property_values": gather_property_values(profile.get("attachment", [])),
    }



import logging
from typing import List

from fastapi import FastAPI
from faststream.rabbit import RabbitBroker
from contextlib import asynccontextmanager

from cattle_grid.testing.accounts import create_test_accounts

from .activity_pub.server import router as ap_router
from .auth import auth_router

from .config import default_filenames
from .account.server import router as fe_router
from .account.rabbit import rabbit_router
from .extensions.load import add_routes_to_api, lifespan_from_extensions
from .dependencies.globals import global_container

from .exchange.exception import exception_middleware
from .app import init_extensions, add_routers_to_broker

from .version import __version__
from .database import upgrade_sql_alchemy

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


tags_description = [
    {
        "name": "activity_pub",
        "description": "Endpoints used and consumed by other Fediverse applications to communicate through cattle_grid",
    },
    {
        "name": "auth",
        "description": """Authentication endpoints
    
The auth endpoint allows one to check the HTTP Signature
and reject requests with an invalid one, only based on the
headers. This step then occurs before the request is passed
to the application. Furthermore, this behavior can be shared
accross many services.""",
    },
]


def create_app(filenames: List[str] = default_filenames) -> FastAPI:
    logger.info("Running cattle grid version %s", __version__)

    global_container.load_config(filenames)

    logger.info("Configuration loaded")

    import os

    alembic_config = __file__.replace("__init__.py", "alembic.ini")

    os.system(f"python -malembic -c {alembic_config} upgrade head")

    if global_container.config.testing.enable:  # type:ignore
        logger.warning("Running in testing mode")

    extensions = init_extensions(global_container.config)

    @asynccontextmanager
    async def lifespan(app: FastAPI, broker: RabbitBroker | None = None):
        if broker is None:
            broker = global_container.broker

        broker.add_middleware(exception_middleware)

        async with global_container.common_lifecycle(global_container.config):  # type:ignore
            await upgrade_sql_alchemy(global_container.engine)

            async with lifespan_from_extensions(extensions):
                if global_container.config.processor_in_app:
                    add_routers_to_broker(broker, extensions, global_container.config)
                session = global_container.async_session_maker()  # type: ignore
                await create_test_accounts(session)

                await broker.start()

                yield
                await broker.stop()

    app = FastAPI(
        lifespan=lifespan,
        title="cattle_grid",
        description="middle ware for the Fediverse",
        version=__version__,
        openapi_tags=tags_description,
    )

    app.include_router(ap_router)
    app.include_router(
        auth_router,
        prefix="/auth",
    )

    app.include_router(fe_router, prefix="/fe")
    app.include_router(rabbit_router)

    add_routes_to_api(app, extensions)

    @app.get("/")
    async def main() -> str:
        return "cattle_grid"

    return app



import click
import tomli_w
from .util import blocklist_form_url_or_file


def save_block_list(block_list, filename="cattle_grid_block_list.toml"):
    with open(filename, "wb") as fp:
        tomli_w.dump({"domain_block": list(block_list)}, fp)


def add_block_command(main):
    @main.group()
    @click.pass_context
    def block(ctx):
        """Allows management of the level 0 blocklist of cattle_grid.
        These are the instances that you consider so bad you want them
        blocked on the lowest level possible."""
        if "config" not in ctx.obj:
            print("Could not load config and it is necessary for blocklist management")
            exit(1)

    @block.command()  # type:ignore
    @click.pass_context
    def empty(ctx):
        """Sets the blocklist to the empty set"""
        ctx.obj["config"].domain_blocks = set()

        save_block_list(set(), filename=ctx.obj["config_file_block_list"])

        exit(0)

    @block.command("list")  # type:ignore
    @click.pass_context
    def list_blocks(ctx):
        """Prints the current blockllist in alphabetical order"""
        blocks = ctx.obj["config"].domain_blocks
        blocks = sorted(blocks)

        for x in blocks:
            print(x)

    @block.command()  # type:ignore
    @click.argument("domain")
    @click.pass_context
    def add(ctx, domain):
        """Adds domain to the current list of domain blocks"""
        ctx.obj["config"].auth.domain_blocks = ctx.obj["config"].auth.domain_blocks | {
            domain
        }
        save_block_list(set(), filename=ctx.obj["config_file_block_list"])

    @block.command()  # type:ignore
    @click.argument("domain")
    @click.pass_context
    def remove(ctx, domain):
        """Removes domain from the current list of domain blocks"""
        ctx.obj["config"].auth.domain_blocks = ctx.obj["config"].auth.domain_blocks - {
            domain
        }
        save_block_list(set(), filename=ctx.obj["config_file_block_list"])

    @block.command()  # type:ignore
    @click.option(
        "--source",
        help="URL or file to import blocklist from",
        default="https://seirdy.one/pb/FediNuke.txt",
    )
    @click.pass_context
    def compare(ctx, source):
        """Compares the current block list with the one provided by source"""
        blocks = blocklist_form_url_or_file(source)
        current = ctx.obj["config"].auth.domain_blocks

        print(f"New blocks in source ({source}):")
        for x in blocks - current:
            print(x)
        print()
        print("Blocks not in source:")
        for x in current - blocks:
            print(x)
        print()

    @block.command("import")  # type:ignore
    @click.option(
        "--source",
        help="URL or file to import blocklist from",
        default="https://seirdy.one/pb/FediNuke.txt",
    )
    @click.option(
        "--overwrite",
        help="Use to overwrite the existing blocklist instead of just adding to it",
        is_flag=True,
        default=False,
    )
    @click.pass_context
    def import_from_source(ctx, source, overwrite):
        """Imports the blocklist from source. By default the blocklist is merged with current one. By specifying the allow flag, one can overwrite the current blocklist."""
        blocks = blocklist_form_url_or_file(source)

        if overwrite:
            ctx.obj["config"].auth.domain_blocks = blocks
        else:
            ctx.obj["config"].auth.domain_blocks = (
                blocks | ctx.obj["config"].auth.domain_blocks
            )
        save_block_list(set(), filename=ctx.obj["config_file_block_list"])

    return block



"""

Sample configuration

```toml
[[extensions]]
module = "cattle_grid.extensions.examples.simple_register"
api_prefix = "/simple_register"

[[extensions.config.registration_types]]
name = "dev"
permissions = ["dev"]
extra_parameters = ["fediverse"]
```

"""

from fastapi import HTTPException

from cattle_grid.account.account import create_account, AccountAlreadyExists
from cattle_grid.dependencies.fastapi import SqlSession
from cattle_grid.extensions import Extension


from .config import RegisterConfiguration

extension = Extension("simple_register", __name__, config_class=RegisterConfiguration)
"""Definition of the extension"""


def determine_registration_type(config: RegisterConfiguration, name):
    for registration_type in config.registration_types:
        if registration_type.name == name:
            return registration_type
    return None


@extension.post("/register/{name}", status_code=201)
async def post_register(
    name,
    body: dict[str, str],
    config: extension.ConfigFastAPI,  # type: ignore
    session: SqlSession,
):
    registration_type = determine_registration_type(config, name)

    if registration_type is None:
        raise HTTPException(404)

    expected_keys = ["name", "password"] + registration_type.extra_parameters

    if any(x not in body for x in expected_keys):
        raise HTTPException(422)
    if any(not isinstance(body[x], str) for x in expected_keys):
        raise HTTPException(422)

    try:
        await create_account(
            session=session,
            name=body["name"],
            password=body["password"],
            permissions=registration_type.permissions,
            meta_information={
                key: body[key] for key in registration_type.extra_parameters
            },
        )
    except AccountAlreadyExists:
        raise HTTPException(409)



"""
Made as a tool to create reports from test runs
done using behave. Basically, a scenario is run, and a markdown file is created.

Currently recorded are the step, the scenario, messages
received on the `incoming.#` and `outgoing.#` routing keys
of the cattle_grid exchange.

Finally, fetch results from gateway are reported.
"""

import logging
import os
import asyncio
import json
from contextlib import asynccontextmanager

from faststream import Context
from faststream.rabbit import RabbitExchange, ExchangeType, RabbitRouter, RabbitQueue

from cattle_grid.dependencies.globals import global_container

logger = logging.getLogger(__name__)

exchange = RabbitExchange("reporting", type=ExchangeType.TOPIC)
lock = asyncio.Lock()
current_file: str | None = None
step_title: str | None = None


def create_reporting_router() -> RabbitRouter:
    router = RabbitRouter()

    @asynccontextmanager
    async def current_file_handler(mode="a"):
        if current_file is None:
            raise ValueError("No current file")
        async with lock:
            with open(current_file, mode) as fp:
                yield fp

    @router.subscriber(RabbitQueue("step_queue", routing_key="step"), exchange)
    async def reporting_step(msg):
        """Reports the current step"""
        if not current_file:
            return

        global step_title

        message_type = msg.get("type")

        step_title = "## " + message_type + ": " + msg.get("name") + "\n\n"

        if message_type in ["when", "then"]:
            async with current_file_handler() as fp:
                fp.writelines([step_title])
                step_title = None

    @router.subscriber(RabbitQueue("scenario_queue", routing_key="scenario"), exchange)
    async def reporting_scenario(msg):
        """Reports the scenario.

        Note every scenario is written to its own file in the `reports` directory."""
        scenario = msg.get("name")
        scenario_filename = "_".join(msg.get("file").split("/")[1:]).removesuffix(
            ".feature"
        )

        async with lock:
            global current_file
            scenario_alpha = "".join([x for x in scenario if x.isalpha()])

            directory = f"reports/{scenario_filename}"
            os.makedirs(directory, exist_ok=True)

            current_file = f"{directory}/{scenario_alpha}.md"

        async with current_file_handler(mode="w") as fp:
            logger.info(msg.get("description"))
            fp.writelines([f"#{scenario}\n\n"] + msg.get("description") + ["\n\n"])

    @router.subscriber(
        RabbitQueue("scenario_end_queue", routing_key="scenario_end"), exchange
    )
    async def reporting_scenario_end():
        """Reports the scenario.

        Note every scenario is written to its own file in the `reports` directory."""

        async with lock:
            global current_file

            current_file = None

    @router.subscriber(
        RabbitQueue("processing_reporting_in", routing_key="receive.#"),
        global_container.account_exchange,
    )
    @router.subscriber(
        RabbitQueue("processing_reporting_out", routing_key="send.#"),
        global_container.account_exchange,
    )
    async def reporting_incoming_outgoing(
        msg: dict,
        routing_key=Context("message.raw_message.routing_key"),
    ):
        """Records incoming and outgoing messages"""
        if not current_file:
            return
        if routing_key.endswith("info"):
            return

        global step_title

        async with current_file_handler() as fp:
            if step_title:
                fp.writelines([step_title])
                step_title = None

            fp.writelines(
                [
                    f"""

```json title="{routing_key}"
""",
                    json.dumps(msg, indent=2),
                    "\n```\n\n",
                ]
            )

        return {}

    return router



from behave import when, then


@when('The request URL is the actor object of "{alice}"')  # type: ignore
def request_url_is_actor(context, alice):
    alice_obj = context.actors[alice]
    context.request = {"url": alice_obj.get("id")}


@when('The request requests the content-type "{content_type}"')  # type: ignore
def request_set_content_type(context, content_type):
    context.request["content_type"] = content_type


@when("The request is made")  # type: ignore
async def make_request(context):
    async with context.session.get(
        context.request["url"],
        headers={"accept": context.request["content_type"]},
    ) as response:
        context.response = response


@then('The response code is "{status_code}"')  # type: ignore
def check_response_status_code(context, status_code):
    assert context.response.status == int(status_code)


@then("The response is a webpage")
def response_is_webpage(context):
    print(context.response)

    headers = context.response.headers

    assert headers.get("content-type").startswith("text/html"), (
        f"Got content-type {headers.get('content-type')}"
    )



import pytest
import asyncio

from .server_sent_events import async_iterator_for_queue_and_task


@pytest.fixture
async def do_nothing_task():
    async def task_function(): ...

    return asyncio.create_task(task_function())


async def test_async_iterator_for_queue_and_task_stops_on_None():
    queue = asyncio.Queue()

    async def task_function():
        await queue.put(None)
        await asyncio.sleep(60)

    task = asyncio.create_task(task_function())

    iterator = async_iterator_for_queue_and_task(queue, task)

    async for _ in iterator:
        ...


async def test_async_iterator_for_queue_and_task_iterators(do_nothing_task):
    queue = asyncio.Queue()
    for x in ["one", "two", "three", None]:
        await queue.put(x)

    iterator = async_iterator_for_queue_and_task(queue, do_nothing_task)

    results = []

    async for x in iterator:
        results.append(x)

    assert results == ["one", "two", "three"]


async def test_heart_beat(do_nothing_task):
    queue = asyncio.Queue()

    iterator = async_iterator_for_queue_and_task(queue, do_nothing_task, timeout=0.01)

    result = await iterator.__anext__()

    assert result == ": <3"



import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from sqlalchemy import select

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.database.account import Account

from . import extension


@pytest.fixture(autouse=True)
def configure_extension():
    extension.configure(
        {
            "registration_types": [
                {
                    "name": "test_name",
                    "permissions": ["permission_one", "permission_two"],
                    "extra_parameters": ["fediverse"],
                }
            ]
        }
    )


@pytest.fixture
def test_client():
    app = FastAPI()
    app.include_router(extension.api_router)

    return TestClient(app)


def test_register_not_found(test_client):
    response = test_client.post(
        "/register/not_found",
        json={"password": "secret", "name": "alice"},
    )

    assert response.status_code == 404


async def test_register(test_client, sql_session_for_test):
    response = test_client.post(
        "/register/test_name",
        json={
            "password": "secret",
            "name": "alice",
            "fediverse": "acct:alice@host.test",
        },
    )

    assert response.status_code == 201

    account = await sql_session_for_test.scalar(
        select(Account).where(Account.name == "alice")
    )
    assert account

    assert account.meta_information == {
        "fediverse": "acct:alice@host.test",
    }
    await sql_session_for_test.refresh(account, attribute_names=["permissions"])
    assert len(account.permissions) == 2


@pytest.mark.parametrize(
    "body",
    [
        {
            "password": "secret",
            "fediverse": "acct:alice@host.test",
        },
        {
            "password": "secret",
            "name": "alice",
        },
        {"password": "secret", "name": "alice", "fediverse": {"object": True}},
    ],
)
def test_register_invalid(test_client, body):
    response = test_client.post(
        "/register/test_name",
        json=body,
    )

    assert response.status_code == 422


def test_register_duplicate(test_client):
    test_client.post(
        "/register/test_name",
        json={
            "password": "secret",
            "name": "alice",
            "fediverse": "acct:alice@host.test",
        },
    )
    response = test_client.post(
        "/register/test_name",
        json={
            "password": "secret",
            "name": "alice",
            "fediverse": "acct:alice@host.test",
        },
    )

    assert response.status_code == 409



from .exchange import UpdateActorMessage, UpdateAction


def test_update_actor_message():
    msg = {
        "actor": "http://local.test/actor",
        "actions": [
            {
                "action": "add_identifier",
                "identifier": "acct:test@example.com",
                "primary": True,
            }
        ],
    }
    data = UpdateActorMessage.model_validate(msg)

    assert isinstance(data.actions[0], UpdateAction)



import pytest

from cattle_grid.testing import mocked_config
from cattle_grid.testing.fixtures import *  # noqa

from .account import create_account, add_permission

from .permissions import allowed_base_urls, can_create_actor_at_base_url

settings_one = {"permissions": {"test_permission": {"base_urls": ["http://one.test"]}}}

settings_admin = {"frontend": {"base_urls": ["http://one.test"]}}


@pytest.fixture
async def test_account(sql_session_for_test):
    account = await create_account(
        sql_session_for_test, "test_account", "test_password"
    )
    assert account
    await add_permission(sql_session_for_test, account, "test_permission")

    return account


@pytest.fixture
async def test_admin_account(sql_session_for_test):
    account = await create_account(
        sql_session_for_test, "test_account", "test_password"
    )
    assert account

    await add_permission(sql_session_for_test, account, "admin")

    return account


async def test_allowed_base_urls_empty(sql_session_for_test, test_account):
    with mocked_config({}):
        result = await allowed_base_urls(sql_session_for_test, test_account)

    assert result == []


async def test_allowed_base_urls(sql_session_for_test, test_account):
    with mocked_config(settings_one):
        result = await allowed_base_urls(
            sql_session_for_test,
            test_account,
        )

    assert result == ["http://one.test"]


async def test_allowed_base_urls_admin(sql_session_for_test, test_admin_account):
    with mocked_config(settings_admin):
        result = await allowed_base_urls(
            sql_session_for_test,
            test_admin_account,
        )

    assert result == ["http://one.test"]


async def test_can_create_actor_at_base_url(sql_session_for_test, test_account):
    with mocked_config({}):
        result = await can_create_actor_at_base_url(
            sql_session_for_test,
            test_account,
            "http://one.test",
        )

    assert not result


async def test_can_create_actor_at_base_url_success(sql_session_for_test, test_account):
    with mocked_config(settings_one):
        result = await can_create_actor_at_base_url(
            sql_session_for_test,
            test_account,
            "http://one.test",
        )

    assert result



import logging


from contextlib import contextmanager
from dynaconf import Dynaconf
from dynaconf.utils import DynaconfDict

from cattle_grid.dependencies.globals import global_container

logger = logging.getLogger(__name__)


@contextmanager
def mocked_config(config: Dynaconf | dict):
    """overrides the configuration stored in `global_container.config`
    with the value in dict, afterwards resets the original value."""
    if isinstance(config, dict):
        config = DynaconfDict(config)
    old_config = global_container.config

    global_container._config = config

    yield

    global_container._config = old_config



import asyncio
from logging.config import fileConfig

from sqlalchemy import pool
from sqlalchemy.engine import Connection
from sqlalchemy.ext.asyncio import async_engine_from_config

from alembic import context

from cattle_grid.database.auth import Base
from cattle_grid.database.activity_pub import Base as APBase

import tomllib

try:
    # this is the Alembic Config object, which provides
    # access to the values within the .ini file in use.
    config = context.config

    # Interpret the config file for Python logging.
    # This line sets up loggers basically.
    if config.config_file_name is not None:
        fileConfig(config.config_file_name)

    # add your model's MetaData object here
    # for 'autogenerate' support
    # from myapp import mymodel
    # target_metadata = mymodel.Base.metadata
except Exception:
    ...

target_metadata = [Base.metadata, APBase.metadata]

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def include_name(name, type_, parent_names):
    if type_ == "table":
        return name.startswith("ORM_PREFIX")
    else:
        return True


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """

    with open("cattle_grid.toml", "rb") as fp:
        data = tomllib.load(fp)
        url = data["db_uri"]

    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        include_name=include_name,
    )

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection: Connection) -> None:
    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        include_name=include_name,
    )

    with context.begin_transaction():
        context.run_migrations()


async def run_async_migrations() -> None:
    """In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    with open("cattle_grid.toml", "rb") as fp:
        data = tomllib.load(fp)
        url = data["db_uri"]

    options = config.get_section(config.config_ini_section, {})
    # print(options)
    options["sqlalchemy.url"] = url

    connectable = async_engine_from_config(
        options,
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    async with connectable.connect() as connection:
        await connection.run_sync(do_run_migrations)

    await connectable.dispose()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode."""

    asyncio.run(run_async_migrations())


try:
    if context.is_offline_mode():
        run_migrations_offline()
    else:
        run_migrations_online()
except Exception:
    ...



import logging
from typing import Sequence

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.database.activity_pub_actor import Actor, Blocking, Follower, Following

logger = logging.getLogger(__name__)


async def is_blocked(session: AsyncSession, actor: Actor, remote_actor_id: str) -> bool:
    """Checks if remote_actor_id is blocked by actor"""
    result = await session.scalar(
        select(Blocking)
        .where(Blocking.active)
        .where(Blocking.actor == actor)
        .where(Blocking.blocking == remote_actor_id)
        .limit(1)
    )

    return bool(result)


async def followers_for_actor(session: AsyncSession, actor: Actor) -> Sequence[str]:
    """Returns the list of accepted followers

    :param actor:
    :returns:
    """

    result = await session.scalars(
        select(Follower.follower)
        .where(Follower.actor == actor)
        .where(Follower.accepted)
    )

    return result.all()


async def following_for_actor(session: AsyncSession, actor: Actor) -> Sequence[str]:
    """Returns the list of accepted people to follow said actor.
    This is the following table.

    :param actor:
    :returns:
    """

    result = await session.scalars(
        select(Following.following)
        .where(Following.actor == actor)
        .where(Following.accepted)
    )

    return result.all()



import logging

from behave import when, then
from sqlalchemy import select
from sqlalchemy.ext.asyncio import async_sessionmaker
from bovine.activitystreams import factories_for_actor_object

from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.activity_pub.actor.requester import is_valid_requester

from cattle_grid.dependencies.globals import global_container

logger = logging.getLogger(__name__)


@when('"{alice}" creates an object addressed to "{recipient}"')
def object_addressed_to(context, alice, recipient):
    alice_actor = context.actors[alice]
    _, object_factory = factories_for_actor_object(alice_actor)

    if recipient == "public":
        context.object = object_factory.note(content="moo").as_public().build()
    elif recipient == "followers":
        context.object = object_factory.note(content="moo").as_followers().build()
    else:
        context.object = object_factory.note(
            content="moo", to={context.actors[recipient].get("id")}
        ).build()


@then('"{bob}" is "{state}" to view this object')
async def check_allowed(context, bob, state):
    bob_id = context.actors[bob].get("id")

    async with global_container.alchemy_database() as engine:
        async with async_sessionmaker(engine)() as session:
            alice = await session.scalar(
                select(Actor).where(
                    Actor.actor_id == context.object.get("attributedTo")
                )
            )
            assert alice
            is_valid = await is_valid_requester(session, bob_id, alice, context.object)

    if is_valid:
        assert state == "authorized"
    else:
        assert state == "unauthorized"



import pytest

from unittest.mock import AsyncMock

from fastapi import FastAPI
from fastapi.testclient import TestClient

from faststream.rabbit import RabbitBroker


from cattle_grid.account.account import create_account
from cattle_grid.dependencies.globals import global_container
from cattle_grid.testing.fixtures import *  # noqa

from . import router


@pytest.fixture
def test_app(sql_session_for_test):
    app = FastAPI()
    app.include_router(router)
    return app


@pytest.fixture
def test_client(test_app):
    return TestClient(test_app)


@pytest.fixture
async def test_token(sql_session_for_test, test_client):
    account = await create_account(sql_session_for_test, "name", "pass")
    assert account
    response = test_client.post(
        "/signin", json={"name": account.name, "password": "pass"}
    )

    assert response.status_code == 200

    return response.json().get("token")


@pytest.fixture
def bearer_header(test_token):
    return {"authorization": f"Bearer {test_token}"}


@pytest.fixture
def test_broker(test_app):
    my_broker = AsyncMock(RabbitBroker)
    test_app.dependency_overrides[global_container.get_broker] = lambda: my_broker

    return my_broker



from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.database.account import ActorForAccount
from cattle_grid.model.exchange_update_actor import UpdateAction, UpdateActionType

from .actor_update import handle_actor_action


async def test_handle_actor_action_rename(
    actor_for_test, account_for_test, sql_session_for_test
):
    actor_for_account = ActorForAccount(
        actor=actor_for_test.actor_id, account=account_for_test
    )
    sql_session_for_test.add(actor_for_account)
    await sql_session_for_test.commit()

    action = UpdateAction(
        action=UpdateActionType.rename,
        name="new name",  # type:ignore
    )

    await handle_actor_action(actor_for_test, sql_session_for_test, action)

    await sql_session_for_test.refresh(actor_for_account)

    assert actor_for_account.name == "new name"



import asyncio
import logging
from faststream import Context
from faststream.rabbit import RabbitRouter, RabbitBroker, RabbitQueue

from cattle_grid.activity_pub.actor import (
    create_actor,
    actor_to_object,
)

from cattle_grid.model.account import (
    FetchMessage,
    FetchResponse,
    InformationResponse,
    CreateActorRequest,
)
from cattle_grid.account.permissions import can_create_actor_at_base_url
from cattle_grid.database.account import ActorForAccount
from cattle_grid.dependencies import CorrelationId, MethodInformation, SqlSession
from cattle_grid.dependencies.globals import global_container

from .info import create_information_response
from .annotations import (
    AccountName,
    AccountFromRoutingKey,
    ActorFromMessage,
)
from .exception import exception_middleware
from .trigger import handle_trigger

logger = logging.getLogger(__name__)


async def handle_fetch(
    msg: FetchMessage,
    correlation_id: CorrelationId,
    name: AccountName,
    actor: ActorFromMessage,
    broker: RabbitBroker = Context(),
):
    """Used to retrieve an object"""
    try:
        async with asyncio.timeout(0.5):
            result = await broker.publish(
                {"actor": actor.actor_id, "uri": msg.uri},
                routing_key="fetch_object",
                exchange=global_container.internal_exchange,
                rpc=True,
            )
        if result == b"":
            result = None
        logger.info("GOT result %s", result)
    except TimeoutError as e:
        logger.error("Request ran into timeout %s", e)
        result = None

    if result:
        await broker.publish(
            FetchResponse(
                uri=msg.uri,
                actor=actor.actor_id,
                data=result,
            ),
            routing_key=f"receive.{name}.response.fetch",
            exchange=global_container.account_exchange,
            correlation_id=correlation_id,
        )
    else:
        await broker.publish(
            {
                "message": "Could not fetch object",
            },
            routing_key=f"error.{name}",
            exchange=global_container.account_exchange,
            correlation_id=correlation_id,
        )


async def create_actor_handler(
    create_message: CreateActorRequest,
    correlation_id: CorrelationId,
    account: AccountFromRoutingKey,
    session: SqlSession,
    broker: RabbitBroker = Context(),
) -> None:
    """Creates an actor associated with the account.

    Updating and deleting actors is done through trigger events."""

    if not await can_create_actor_at_base_url(
        session, account, create_message.base_url
    ):
        raise ValueError(f"Base URL {create_message.base_url} not in allowed base urls")

    actor = await create_actor(
        session,
        create_message.base_url,
        preferred_username=create_message.preferred_username,
        profile=create_message.profile,
    )

    session.add(
        ActorForAccount(
            account=account,
            actor=actor.actor_id,
            name=create_message.name or "from drive",
        )
    )

    if create_message.automatically_accept_followers:
        actor.automatically_accept_followers = True

    result = actor_to_object(actor)

    await session.refresh(account)

    logger.info("Created actor %s for %s", actor.actor_id, account.name)

    await broker.publish(
        result,
        routing_key=f"receive.{account.name}.response.create_actor",
        exchange=global_container.account_exchange,
        correlation_id=correlation_id,
    )
    await session.commit()


def create_router() -> RabbitRouter:
    router = RabbitRouter(middlewares=[exception_middleware])

    info_publisher = router.publisher(
        "receive.name.response.fetch",
        schema=FetchResponse,
        title="receive.NAME.response.fetch",
    )

    router.subscriber(
        RabbitQueue(
            "send_request_fetch",
            routing_key="send.*.request.fetch",
        ),
        exchange=global_container.account_exchange,
        title="send.*.request.fetch",
    )(handle_fetch)

    info_publisher = router.publisher(
        "receive.name.response.info",
        schema=InformationResponse,
        title="receive.NAME.response.info",
    )

    @router.subscriber(
        RabbitQueue(
            "send_request_info",
            routing_key="send.*.request.info",
        ),
        exchange=global_container.account_exchange,
        title="send.*.request.info",
    )
    async def information_request_handler(
        account: AccountFromRoutingKey,
        correlation_id: CorrelationId,
        method_information: MethodInformation,
        session: SqlSession,
        broker: RabbitBroker = Context(),
    ) -> None:
        """Provides information about the underlying service"""
        await info_publisher.publish(
            await create_information_response(session, account, method_information),
            routing_key=f"receive.{account.name}.response.info",
            exchange=global_container.account_exchange,
            correlation_id=correlation_id,
        )

    info_publisher = router.publisher(
        "receive.name.response.create_actor",
        schema=dict,
        title="receive.NAME.response.create_actor",
        description="""The response to a create_actor request
    is the actor profile (as formatted towards the Fediverse)
    This might change in the future""",
    )

    router.subscriber(
        RabbitQueue(
            "send_request_create_actor",
            routing_key="send.*.request.create_actor",
        ),
        exchange=global_container.account_exchange,
        title="send.*.request.create_actor",
    )(create_actor_handler)

    router.subscriber(
        RabbitQueue(
            "send_trigger",
            routing_key="send.*.trigger.#",
        ),
        exchange=global_container.account_exchange,
        title="send.*.trigger.#",
    )(handle_trigger)

    return router



from faststream.rabbit import RabbitBroker, RabbitExchange
from cattle_grid.model import ActivityMessage, SharedInboxMessage


def determine_activity_type(activity: dict) -> str | None:
    """Determines the type of an activity

    ```pycon
    >>> determine_activity_type({"type": "Follow"})
    'Follow'

    ```


    ```pycon
    >>> determine_activity_type({}) is None
    True

    ```

    In the case of multiple types, these are concatenated. This means
    that they are probably missed by processing, but don't get ignored.

    ```pycon
    >>> determine_activity_type({"type": ["Follow", "WhileSkipping"]})
    'FollowWhileSkipping'

    ```

    :params activity:
    :returns:

    """

    activity_type = activity.get("type")
    if activity_type is None:
        return None
    if isinstance(activity_type, list):
        activity_type = "".join(activity_type)

    return activity_type


async def enqueue_from_inbox(
    broker: RabbitBroker,
    exchange: RabbitExchange,
    receiving_actor_id: str,
    content: dict,
):
    """Enqueues a new message arrived from the inbox

    The routing key will be `incoming.${activity_type}`
    """
    activity_type = determine_activity_type(content)
    if activity_type is None:
        return

    msg = ActivityMessage(actor=receiving_actor_id, data=content)

    await broker.publish(
        msg, exchange=exchange, routing_key=f"incoming.{activity_type}"
    )


async def enqueue_from_shared_inbox(
    broker: RabbitBroker,
    exchange: RabbitExchange,
    content: dict,
):
    """Enqueues a new message arrived from the inbox

    The routing key will be `incoming.${activity_type}`
    """
    activity_type = determine_activity_type(content)
    if activity_type is None:
        return

    msg = SharedInboxMessage(data=content)

    await broker.publish(msg, exchange=exchange, routing_key="shared_inbox")



import pytest
from unittest.mock import AsyncMock
from sqlalchemy import select
from cattle_grid.account.account import add_permission
from cattle_grid.database.account import Account, ActorForAccount

from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.model.account import CreateActorRequest
from .router import create_actor_handler


async def test_create_actor_handler_no_permission(sql_session_for_test):
    account = Account(name="test", password_hash="")
    sql_session_for_test.add(account)
    await sql_session_for_test.commit()
    broker = AsyncMock()

    with pytest.raises(ValueError):
        await create_actor_handler(
            CreateActorRequest(base_url="http://abel", preferred_username="username"),
            account=account,
            broker=broker,
            correlation_id="uuid",
            session=sql_session_for_test,
        )


async def test_create_actor_handler(sql_session_for_test):
    account = Account(name="test", password_hash="")
    sql_session_for_test.add(account)
    await sql_session_for_test.commit()
    await add_permission(sql_session_for_test, account, "admin")
    broker = AsyncMock()

    await create_actor_handler(
        CreateActorRequest(base_url="http://abel", preferred_username="username"),
        account=account,
        broker=broker,
        correlation_id="uuid",
        session=sql_session_for_test,
    )

    result = [x for x in await sql_session_for_test.scalars(select(ActorForAccount))]

    assert len(result) == 1

    broker.publish.assert_awaited_once()

    (data,) = broker.publish.call_args[0]

    assert data["id"] == result[0].actor
    assert data["preferredUsername"] == "username"



from click.testing import CliRunner

from cattle_grid.testing.cli import *  # noqa

from .__main__ import main


def test_list(db_uri, create_database):
    runner = CliRunner(env={"CATTLE_GRID_DB_URI": db_uri})
    result = runner.invoke(main, ["list"])

    assert result.exit_code == 0



import asyncio

from unittest.mock import AsyncMock, MagicMock
from contextlib import asynccontextmanager

import pytest

from cattle_grid.model.account import EventType

from .streaming import get_message_streamer, construct_routing_key


async def test_get_message_streamer():
    broker = MagicMock()

    @asynccontextmanager
    async def connection():
        yield AsyncMock()

    broker._connection = connection()

    streamer = get_message_streamer(AsyncMock(), 0.1)

    result = streamer("account_name", EventType.incoming)

    assert isinstance(result[0], asyncio.Queue)
    assert isinstance(result[1], asyncio.Task)

    try:
        result[1].cancel()
    except Exception:
        ...


@pytest.mark.parametrize(
    "event_type,expected",
    [
        (EventType.incoming, "receive.alice.incoming"),
        (EventType.outgoing, "receive.alice.outgoing"),
        (EventType.error, "error.alice"),
    ],
)
def test_construct_routing_key(event_type, expected):
    result = construct_routing_key("alice", event_type)

    assert result == expected
